"""Integration tests for the IpcsCommand."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

import pebble_shell.commands
import pebble_shell.shell

if TYPE_CHECKING:
    import ops


@pytest.fixture
def command(shell: pebble_shell.shell.PebbleShell):
    """Fixture to create a IpcsCommand instance."""
    yield pebble_shell.commands.IpcsCommand(shell=shell)


def test_name(command: pebble_shell.commands.IpcsCommand):
    assert command.name == "ipcs"


def test_category(command: pebble_shell.commands.IpcsCommand):
    assert command.category == "System Information"


def test_help(command: pebble_shell.commands.IpcsCommand):
    with command.shell.console.capture() as capture:
        command.show_help()
    output = capture.get()
    assert "ipcs" in output
    assert "Show IPC facilities" in output
    assert "semaphore" in output


@pytest.mark.parametrize("args", [["-h"], ["--help"]])
def test_execute_help(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
    args: list[str],
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=args)
    assert result == 0
    output = capture.get()
    assert "ipcs" in output
    assert "Show IPC facilities" in output


def test_execute_no_args_show_all(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # ipcs with no args should show all IPC facilities
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should succeed showing IPC information
    assert result == 0
    output = capture.get()
    # Should contain IPC facility information
    assert len(output.strip()) >= 0
    # May contain headers for different IPC types
    if len(output.strip()) > 0:
        assert any(
            keyword in output.lower()
            for keyword in ["shared memory", "semaphore", "message", "key", "id"]
        )


def test_execute_show_all_facilities(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test showing all IPC facilities
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-a"])

    # Should succeed showing all facilities
    assert result == 0
    output = capture.get()
    # Should contain comprehensive IPC information
    assert len(output.strip()) >= 0
    # Should show all IPC types
    if len(output.strip()) > 0:
        assert any(
            section in output.lower()
            for section in ["shared memory", "semaphore", "message"]
        )


def test_execute_shared_memory_segments(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test showing shared memory segments
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-m"])

    # Should succeed showing shared memory
    assert result == 0
    output = capture.get()
    # Should contain shared memory information
    assert len(output.strip()) >= 0
    # May contain shared memory headers
    if len(output.strip()) > 0:
        assert any(
            keyword in output.lower()
            for keyword in ["shared memory", "shmid", "key", "size", "owner"]
        )


def test_execute_semaphore_arrays(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test showing semaphore arrays
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-s"])

    # Should succeed showing semaphores
    assert result == 0
    output = capture.get()
    # Should contain semaphore information
    assert len(output.strip()) >= 0
    # May contain semaphore headers
    if len(output.strip()) > 0:
        assert any(
            keyword in output.lower()
            for keyword in ["semaphore", "semid", "key", "nsems", "owner"]
        )


def test_execute_message_queues(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test showing message queues
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-q"])

    # Should succeed showing message queues
    assert result == 0
    output = capture.get()
    # Should contain message queue information
    assert len(output.strip()) >= 0
    # May contain message queue headers
    if len(output.strip()) > 0:
        assert any(
            keyword in output.lower()
            for keyword in ["message", "msqid", "key", "messages", "owner"]
        )


def test_execute_limits_information(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test showing IPC limits
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-l"])

    # Should succeed showing limits
    assert result == 0
    output = capture.get()
    # Should contain limits information
    assert len(output.strip()) > 0
    # Should contain limit values
    assert any(
        keyword in output.lower()
        for keyword in [
            "limit",
            "max",
            "resource",
            "shared memory",
            "semaphore",
            "message",
        ]
    )


def test_execute_summary_information(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test showing summary information
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-u"])

    # Should succeed showing summary
    assert result == 0
    output = capture.get()
    # Should contain summary information
    assert len(output.strip()) >= 0
    # May contain usage statistics
    if len(output.strip()) > 0:
        assert any(
            keyword in output.lower()
            for keyword in ["used", "allocated", "total", "summary"]
        )


def test_execute_time_information(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test showing time information
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-t"])

    # Should succeed showing time info
    assert result == 0
    output = capture.get()
    # Should contain time information
    assert len(output.strip()) >= 0
    # May contain timestamps
    if len(output.strip()) > 0:
        assert any(
            keyword in output.lower()
            for keyword in ["time", "attach", "detach", "change", "create"]
        )


def test_execute_process_information(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test showing process information
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-p"])

    # Should succeed showing process info
    assert result == 0
    output = capture.get()
    # Should contain process information
    assert len(output.strip()) >= 0
    # May contain process IDs
    if len(output.strip()) > 0:
        assert any(
            keyword in output.lower()
            for keyword in ["pid", "process", "creator", "last"]
        )


def test_execute_creator_information(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test showing creator information
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-c"])

    # Should succeed showing creator info
    assert result == 0
    output = capture.get()
    # Should contain creator information
    assert len(output.strip()) >= 0
    # May contain creator details
    if len(output.strip()) > 0:
        assert any(
            keyword in output.lower()
            for keyword in ["creator", "owner", "cuid", "cgid"]
        )


def test_execute_numeric_output(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test numeric output format
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-n"])

    # Should succeed with numeric format
    assert result == 0
    output = capture.get()
    # Should contain numeric information
    assert len(output.strip()) >= 0
    # Should show numeric IDs instead of names
    if len(output.strip()) > 0:
        # May contain numeric UIDs/GIDs
        assert any(char.isdigit() for char in output)


def test_execute_combined_options(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test combined options
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-mpt"])

    # Should succeed with combined options
    assert result == 0
    output = capture.get()
    # Should contain combined information
    assert len(output.strip()) >= 0
    # Should show shared memory with process and time info
    if len(output.strip()) > 0:
        assert any(
            keyword in output.lower() for keyword in ["shared memory", "pid", "time"]
        )


def test_execute_specific_facility_id(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test showing specific facility by ID
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-m", "-i", "999999"])

    # Should either show specific facility or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show specific facility information
        assert len(output.strip()) >= 0
    else:
        # Should fail if facility doesn't exist
        assert result == 1
        output = capture.get()
        assert any(
            msg in output for msg in ["invalid", "not found", "no such", "error"]
        )


def test_execute_human_readable_format(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test human-readable format
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-h", "-m"])

    # Should either show help or human-readable format
    if result == 0:
        output = capture.get()
        # Should contain formatted information
        assert len(output.strip()) >= 0
    else:
        # May not support -h for human-readable
        assert result == 1


def test_execute_invalid_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test invalid option
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-z"])

    # Should fail with invalid option error
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["invalid", "unknown", "option", "usage"])


def test_execute_permission_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test permission handling
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-a"])

    # Should succeed showing accessible facilities
    assert result == 0
    output = capture.get()
    # Should show only accessible IPC facilities
    assert len(output.strip()) >= 0


def test_execute_empty_system_state(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test with empty IPC system state
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-m"])

    # Should succeed even with no shared memory segments
    assert result == 0
    output = capture.get()
    # Should show headers even if no segments exist
    assert len(output.strip()) >= 0


def test_execute_output_format_validation(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test output format validation
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-l"])

    # Should produce properly formatted output
    assert result == 0
    output = capture.get()

    if len(output.strip()) > 0:
        # Should have structured output format
        _ = output.strip().split("\n")
        # Should contain limit information
        assert any(
            keyword in output.lower() for keyword in ["limit", "max", "resource"]
        )


def test_execute_memory_efficiency(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test memory efficiency
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-a"])

    # Should be memory efficient
    assert result == 0
    output = capture.get()
    # Should not consume excessive memory
    assert len(output) < 50000  # Reasonable output size limit


def test_execute_concurrent_execution(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test concurrent execution safety
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-m"])

    # Should handle concurrent execution safely
    assert result == 0
    output = capture.get()
    assert len(output.strip()) >= 0


def test_execute_error_recovery(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test error recovery capabilities
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-i", "invalid"])

    # Should recover from errors gracefully
    assert result == 1
    output = capture.get()
    # Should provide meaningful error message
    assert any(msg in output for msg in ["invalid", "error", "usage"])


def test_execute_system_resource_monitoring(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test system resource monitoring
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-u"])

    # Should monitor system resources appropriately
    assert result == 0
    output = capture.get()
    # Should show resource usage
    assert len(output.strip()) >= 0


def test_execute_ipc_namespace_awareness(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test IPC namespace awareness
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-a"])

    # Should work within current IPC namespace
    assert result == 0
    output = capture.get()
    # Should show facilities in current namespace
    assert len(output.strip()) >= 0


def test_execute_cross_platform_compatibility(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test cross-platform compatibility
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-l"])

    # Should work across different platforms
    assert result == 0
    output = capture.get()
    # Should adapt to platform-specific IPC implementation
    assert len(output.strip()) >= 0


def test_execute_performance_monitoring(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test performance monitoring
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-t"])

    # Should complete efficiently
    assert result == 0
    output = capture.get()
    # Should provide timing information efficiently
    assert len(output.strip()) >= 0


def test_execute_security_awareness(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test security awareness
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-c"])

    # Should handle security information appropriately
    assert result == 0
    output = capture.get()
    # Should show creator/owner information securely
    assert len(output.strip()) >= 0


def test_execute_locale_independence(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test locale independence
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-l"])

    # Should work regardless of locale settings
    assert result == 0
    output = capture.get()
    # Should be locale-independent
    assert len(output.strip()) >= 0


def test_execute_signal_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test signal handling during execution
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-a"])

    # Should handle signals appropriately
    assert result == 0
    output = capture.get()
    # Should be signal-safe
    assert len(output.strip()) >= 0


def test_execute_kernel_interface_compatibility(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test kernel interface compatibility
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-m"])

    # Should interface with kernel IPC subsystem correctly
    assert result == 0
    output = capture.get()
    # Should work with current kernel version
    assert len(output.strip()) >= 0


def test_execute_data_consistency(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test data consistency
    with command.shell.console.capture() as _:
        result1 = command.execute(client=client, args=["-m"])

    with command.shell.console.capture() as _:
        result2 = command.execute(client=client, args=["-m"])

    # Should produce consistent results
    assert result1 == 0
    assert result2 == 0
    # Both executions should succeed consistently
    assert result1 == result2


def test_execute_verbose_output(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test verbose output mode
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-v"])

    # Should either provide verbose output or show error
    if result == 0:
        output = capture.get()
        # Should provide detailed information
        assert len(output.strip()) >= 0
    else:
        # May not support -v option
        assert result == 1


def test_execute_facility_filtering(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test facility filtering
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-s", "-q"])

    # Should show only semaphores and message queues
    assert result == 0
    output = capture.get()
    # Should filter facilities appropriately
    assert len(output.strip()) >= 0


def test_execute_backward_compatibility(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test backward compatibility
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-a"])

    # Should maintain backward compatibility
    assert result == 0
    output = capture.get()
    # Should work with older IPC implementations
    assert len(output.strip()) >= 0


def test_execute_resource_cleanup(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test resource cleanup
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-l"])

    # Should clean up resources properly
    assert result == 0
    output = capture.get()
    # Should not leak resources
    assert len(output.strip()) >= 0


def test_execute_formatting_consistency(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test formatting consistency
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-m"])

    # Should produce consistently formatted output
    assert result == 0
    output = capture.get()

    if len(output.strip()) > 0:
        # Should have consistent column formatting
        lines = output.strip().split("\n")
        # Should maintain formatting consistency
        assert len(lines) >= 1


def test_execute_ipc_state_detection(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test IPC state detection
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-u"])

    # Should detect current IPC state accurately
    assert result == 0
    output = capture.get()
    # Should reflect actual system state
    assert len(output.strip()) >= 0


def test_execute_multi_user_environment(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test multi-user environment handling
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-c"])

    # Should handle multi-user environments appropriately
    assert result == 0
    output = capture.get()
    # Should show ownership information properly
    assert len(output.strip()) >= 0


def test_execute_robust_operation(
    client: ops.pebble.Client,
    command: pebble_shell.commands.IpcsCommand,
):
    # Test robust operation under stress
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-a"])

    # Should operate robustly
    assert result == 0
    output = capture.get()
    # Should handle stress conditions
    assert len(output.strip()) >= 0



"""Implementation of PatchCommand."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING, Union

import ops

from ...utils.command_helpers import handle_help_flag, parse_flags, safe_read_file
from .._base import Command

if TYPE_CHECKING:
    import shimmer


# TODO: Use the prototype from Shimmer.
ClientType = Union[ops.pebble.Client, "shimmer.PebbleCliClient"]


class PatchCommand(Command):
    """Implementation of patch command."""

    name = "patch"
    help = "Apply patch files to remote files"
    category = "Advanced Utilities"

    def show_help(self):
        """Show command help."""
        help_text = """Apply patch files to remote files.

Usage: patch [OPTIONS] [ORIGINAL [PATCHFILE]]

Description:
    Apply a patch file to an original file on the remote system.
    Supports unified diff format.

Options:
    -p NUM          Strip NUM leading path components from file names
    -R, --reverse   Apply patch in reverse
    -o FILE         Output to FILE instead of in-place
    -i PATCHFILE    Use PATCHFILE as patch input
    --dry-run       Show what would be done without making changes
    -h, --help      Show this help message

Examples:
    patch -i changes.patch file.txt
    patch -p1 < changes.patch
    patch --dry-run -i test.patch
        """
        self.console.print(help_text)

    def execute(self, client: ClientType, args: list[str]) -> int:
        """Execute the patch command."""
        if handle_help_flag(self, args):
            return 0

        parse_result = parse_flags(
            args,
            {
                "p": int,  # strip path components
                "R": bool,  # reverse
                "reverse": bool,
                "o": str,  # output file
                "i": str,  # input patch file
                "dry-run": bool,  # dry run
            },
            self.shell,
        )
        if parse_result is None:
            return 1
        flags, positional_args = parse_result

        strip_count = flags.get("p") or 0
        reverse = flags.get("R", False) or flags.get("reverse", False)
        output_file = flags.get("o")
        patch_file = flags.get("i")
        dry_run = flags.get("dry-run", False)

        try:
            # Determine patch source and target files
            if patch_file:
                # Patch file specified with -i
                if len(positional_args) == 0:
                    self.console.print("[red]patch: missing target file[/red]")
                    return 1
                target_file = positional_args[0]
            elif len(positional_args) >= 2:
                # patch original patchfile
                target_file = positional_args[0]
                patch_file = positional_args[1]
            elif len(positional_args) == 1:
                # patch original (patch from stdin - not supported)
                self.console.print("[red]patch: reading from stdin not supported[/red]")
                return 1
            else:
                self.console.print("[red]patch: missing arguments[/red]")
                return 1

            # Read patch file
            patch_content = safe_read_file(client, patch_file)
            if patch_content is None:
                self.console.print(
                    f"[red]patch: cannot read patch file '{patch_file}'[/red]"
                )
                return 1

            # Read target file
            original_content = safe_read_file(client, target_file)
            if original_content is None:
                self.console.print(
                    f"[red]patch: cannot read target file '{target_file}'[/red]"
                )
                return 1

            # Apply patch
            try:
                patched_content = self._apply_patch(
                    original_content, patch_content, reverse, strip_count
                )

                if dry_run:
                    self.console.print(
                        f"[green]patch: would patch '{target_file}'[/green]"
                    )
                    return 0

                # Write result
                output_path = output_file if output_file else target_file
                with client.push(output_path, encoding="utf-8") as f:
                    f.write(patched_content)

                self.console.print(
                    f"[green]patch: successfully patched '{output_path}'[/green]"
                )
                return 0

            except Exception as e:
                self.console.print(f"[red]patch: {e}[/red]")
                return 1

        except Exception as e:
            self.console.print(f"[red]patch: {e}[/red]")
            return 1

    def _apply_patch(
        self, original: str, patch_content: str, reverse: bool, strip_count: int
    ) -> str:
        """Apply a unified diff patch to the original content."""
        lines = original.splitlines(keepends=True)
        patch_lines = patch_content.splitlines()

        # Parse patch hunks
        hunks = self._parse_unified_diff(patch_lines, strip_count)

        if reverse:
            # Reverse the patch operations
            hunks = self._reverse_hunks(hunks)

        # Apply hunks in reverse order to maintain line numbers
        for hunk in reversed(hunks):
            lines = self._apply_hunk(lines, hunk)

        return "".join(lines)

    def _parse_unified_diff(
        self, patch_lines: list[str], strip_count: int
    ) -> list[dict]:
        """Parse unified diff format into hunks."""
        hunks = []
        current_hunk = None

        for line in patch_lines:
            if line.startswith("--- ") or line.startswith("+++ "):
                continue  # Skip file headers
            elif line.startswith("@@"):
                # Parse hunk header: @@ -start,count +start,count @@
                match = re.match(r"@@ -(\d+)(?:,(\d+))? \+(\d+)(?:,(\d+))? @@", line)
                if match:
                    old_start, old_count, new_start, new_count = match.groups()
                    current_hunk = {
                        "old_start": int(old_start),
                        "old_count": int(old_count) if old_count else 1,
                        "new_start": int(new_start),
                        "new_count": int(new_count) if new_count else 1,
                        "lines": [],
                    }
                    hunks.append(current_hunk)
            elif current_hunk is not None and (
                line.startswith(" ") or line.startswith("-") or line.startswith("+")
            ):
                current_hunk["lines"].append(line)

        return hunks

    def _reverse_hunks(self, hunks: list[dict]) -> list[dict]:
        """Reverse patch hunks for reverse application."""
        reversed_hunks = []

        for hunk in hunks:
            reversed_hunk = {
                "old_start": hunk["new_start"],
                "old_count": hunk["new_count"],
                "new_start": hunk["old_start"],
                "new_count": hunk["old_count"],
                "lines": [],
            }

            # Reverse the line operations
            for line in hunk["lines"]:
                if line.startswith("+"):
                    reversed_hunk["lines"].append("-" + line[1:])
                elif line.startswith("-"):
                    reversed_hunk["lines"].append("+" + line[1:])
                else:
                    reversed_hunk["lines"].append(line)

            reversed_hunks.append(reversed_hunk)

        return reversed_hunks

    def _apply_hunk(self, lines: list[str], hunk: dict) -> list[str]:
        """Apply a single hunk to the lines."""
        # Convert to 0-based indexing
        start_line = hunk["old_start"] - 1

        # Extract lines to modify
        old_lines = []
        new_lines = []

        for patch_line in hunk["lines"]:
            if patch_line.startswith(" "):
                # Context line
                old_lines.append(patch_line[1:])
                new_lines.append(patch_line[1:])
            elif patch_line.startswith("-"):
                # Removed line
                old_lines.append(patch_line[1:])
            elif patch_line.startswith("+"):
                # Added line
                new_lines.append(patch_line[1:])

        # Replace the section
        end_line = start_line + len(old_lines)
        return lines[:start_line] + new_lines + lines[end_line:]



"""Integration tests for the RebootCommand."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

import pebble_shell.commands
import pebble_shell.shell

if TYPE_CHECKING:
    import ops


@pytest.fixture
def command(shell: pebble_shell.shell.PebbleShell):
    """Fixture to create a RebootCommand instance."""
    yield pebble_shell.commands.RebootCommand(shell=shell)


def test_name(command: pebble_shell.commands.RebootCommand):
    assert command.name == "pebble-reboot"


def test_category(command: pebble_shell.commands.RebootCommand):
    assert command.category == "Pebble Management"


def test_help(command: pebble_shell.commands.RebootCommand):
    with command.shell.console.capture() as capture:
        command.show_help()
    output = capture.get()
    assert "reboot" in output
    assert "Reboot the system" in output


@pytest.mark.parametrize("args", [["-h"], ["--help"]])
def test_execute_help(
    client: ops.pebble.Client,
    command: pebble_shell.commands.RebootCommand,
    args: list[str],
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=args)
    assert result == 0
    output = capture.get()
    assert "reboot" in output
    assert "Reboot the system" in output


def test_execute_no_args_immediate_reboot(
    client: ops.pebble.Client,
    command: pebble_shell.commands.RebootCommand,
):
    # reboot with no args should initiate immediate reboot
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should either succeed with reboot message or fail if Pebble unavailable
    if result == 0:
        output = capture.get()
        # Should show reboot initiation message
        assert any(
            msg in output
            for msg in ["Reboot initiated", "System will reboot", "Reboot scheduled"]
        )
    else:
        # Should fail if Pebble API unavailable or permission denied
        assert result == 1
        output = capture.get()
        assert any(
            msg in output for msg in ["Reboot operation failed", "error", "failed"]
        )


def test_execute_immediate_reboot_confirmation(
    client: ops.pebble.Client,
    command: pebble_shell.commands.RebootCommand,
):
    # Test immediate reboot confirmation message
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should confirm reboot initiation if successful
    if result == 0:
        output = capture.get()
        assert any(
            msg in output
            for msg in ["Reboot initiated", "System will reboot", "Reboot scheduled"]
        )


def test_execute_api_error_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.RebootCommand,
):
    # Test API error handling when Pebble is unavailable
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should handle API errors gracefully
    if result == 1:
        output = capture.get()
        assert any(
            msg in output for msg in ["Reboot operation failed", "error", "failed"]
        )
    else:
        # Should succeed if Pebble API is available
        assert result == 0


def test_execute_permission_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.RebootCommand,
):
    # Test handling of permission errors for reboot
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should handle permission errors gracefully
    if result == 1:
        output = capture.get()
        # Should show appropriate error message
        assert any(msg in output for msg in ["permission", "denied", "error", "failed"])
    else:
        # May succeed if user has reboot permissions
        assert result == 0


def test_execute_system_state_validation(
    client: ops.pebble.Client,
    command: pebble_shell.commands.RebootCommand,
):
    # Test system state validation before reboot
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should validate system state before proceeding
    if result == 0:
        output = capture.get()
        # Should proceed with reboot if system is in valid state
        assert any(msg in output for msg in ["Reboot initiated", "System will reboot"])
    else:
        # May fail if system is not in valid state for reboot
        assert result == 1


def test_execute_reboot_request_submission(
    client: ops.pebble.Client,
    command: pebble_shell.commands.RebootCommand,
):
    # Test reboot request submission to Pebble
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should submit reboot request to Pebble API
    if result == 0:
        output = capture.get()
        # Should indicate successful submission
        assert len(output.strip()) > 0
    else:
        assert result == 1


def test_execute_extra_args_ignored(
    client: ops.pebble.Client,
    command: pebble_shell.commands.RebootCommand,
):
    # Test that extra arguments are ignored
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["extra", "args", "ignored"])

    # Should ignore extra arguments and proceed with reboot
    if result == 0:
        output = capture.get()
        # Should proceed with reboot regardless of extra args
        assert any(msg in output for msg in ["Reboot initiated", "System will reboot"])
    else:
        assert result == 1


def test_execute_concurrent_reboot_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.RebootCommand,
):
    # Test handling of concurrent reboot requests
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should handle concurrent reboot requests appropriately
    if result == 0:
        output = capture.get()
        # Should either proceed or indicate reboot already in progress
        assert len(output.strip()) > 0
    else:
        assert result == 1


def test_execute_reboot_safety_checks(
    client: ops.pebble.Client,
    command: pebble_shell.commands.RebootCommand,
):
    # Test reboot safety checks
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should perform safety checks before reboot
    if result == 0:
        output = capture.get()
        # Should proceed after safety checks pass
        assert any(msg in output for msg in ["Reboot initiated", "System will reboot"])
    else:
        # May fail if safety checks fail
        assert result == 1


def test_execute_service_shutdown_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.RebootCommand,
):
    # Test handling of service shutdown during reboot
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should handle service shutdown as part of reboot
    if result == 0:
        output = capture.get()
        # Should initiate shutdown sequence
        assert any(msg in output for msg in ["Reboot initiated", "System will reboot"])
    else:
        assert result == 1


def test_execute_graceful_shutdown_sequence(
    client: ops.pebble.Client,
    command: pebble_shell.commands.RebootCommand,
):
    # Test graceful shutdown sequence before reboot
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should initiate graceful shutdown sequence
    if result == 0:
        output = capture.get()
        # Should handle graceful shutdown
        assert any(msg in output for msg in ["Reboot initiated", "System will reboot"])
    else:
        assert result == 1


def test_execute_filesystem_sync_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.RebootCommand,
):
    # Test filesystem sync handling before reboot
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should handle filesystem synchronization
    if result == 0:
        output = capture.get()
        # Should proceed after filesystem sync
        assert any(msg in output for msg in ["Reboot initiated", "System will reboot"])
    else:
        assert result == 1


def test_execute_process_termination_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.RebootCommand,
):
    # Test process termination handling during reboot
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should handle process termination appropriately
    if result == 0:
        output = capture.get()
        # Should manage process termination sequence
        assert any(msg in output for msg in ["Reboot initiated", "System will reboot"])
    else:
        assert result == 1


def test_execute_hardware_reset_preparation(
    client: ops.pebble.Client,
    command: pebble_shell.commands.RebootCommand,
):
    # Test hardware reset preparation
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should prepare for hardware reset
    if result == 0:
        output = capture.get()
        # Should prepare system for hardware reset
        assert any(msg in output for msg in ["Reboot initiated", "System will reboot"])
    else:
        assert result == 1


def test_execute_reboot_logging(
    client: ops.pebble.Client,
    command: pebble_shell.commands.RebootCommand,
):
    # Test reboot event logging
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should log reboot events appropriately
    if result == 0:
        output = capture.get()
        # Should indicate logging of reboot event
        assert len(output.strip()) > 0
    else:
        assert result == 1


def test_execute_emergency_reboot_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.RebootCommand,
):
    # Test emergency reboot handling
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should handle emergency reboot scenarios
    if result == 0:
        output = capture.get()
        # Should proceed with reboot even in emergency
        assert any(msg in output for msg in ["Reboot initiated", "System will reboot"])
    else:
        assert result == 1


def test_execute_container_environment_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.RebootCommand,
):
    # Test reboot in container environment
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should handle reboot appropriately in container context
    if result == 0:
        output = capture.get()
        # Should adapt reboot behavior for container environment
        assert any(
            msg in output
            for msg in ["Reboot initiated", "System will reboot", "Container restart"]
        )
    else:
        assert result == 1


def test_execute_reboot_timing_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.RebootCommand,
):
    # Test reboot timing and scheduling
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should handle reboot timing appropriately
    if result == 0:
        output = capture.get()
        # Should indicate when reboot will occur
        assert any(msg in output for msg in ["Reboot initiated", "System will reboot"])
    else:
        assert result == 1


def test_execute_network_cleanup_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.RebootCommand,
):
    # Test network cleanup during reboot
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should handle network cleanup appropriately
    if result == 0:
        output = capture.get()
        # Should manage network cleanup sequence
        assert any(msg in output for msg in ["Reboot initiated", "System will reboot"])
    else:
        assert result == 1


def test_execute_resource_cleanup_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.RebootCommand,
):
    # Test resource cleanup during reboot
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should handle resource cleanup appropriately
    if result == 0:
        output = capture.get()
        # Should clean up resources before reboot
        assert any(msg in output for msg in ["Reboot initiated", "System will reboot"])
    else:
        assert result == 1


def test_execute_reboot_confirmation_message(
    client: ops.pebble.Client,
    command: pebble_shell.commands.RebootCommand,
):
    # Test reboot confirmation message format
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should provide clear confirmation message
    if result == 0:
        output = capture.get()
        # Should have informative confirmation message
        assert len(output.strip()) > 0
        assert any(
            msg in output
            for msg in ["Reboot initiated", "System will reboot", "reboot"]
        )


def test_execute_pebble_integration_validation(
    client: ops.pebble.Client,
    command: pebble_shell.commands.RebootCommand,
):
    # Test Pebble integration validation
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should validate Pebble integration before reboot
    if result == 0:
        output = capture.get()
        # Should proceed after validation
        assert any(msg in output for msg in ["Reboot initiated", "System will reboot"])
    else:
        # May fail if Pebble integration validation fails
        assert result == 1


def test_execute_system_health_check(
    client: ops.pebble.Client,
    command: pebble_shell.commands.RebootCommand,
):
    # Test system health check before reboot
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should perform system health check
    if result == 0:
        output = capture.get()
        # Should proceed after health check
        assert any(msg in output for msg in ["Reboot initiated", "System will reboot"])
    else:
        # May fail if health check fails
        assert result == 1


def test_execute_reboot_state_management(
    client: ops.pebble.Client,
    command: pebble_shell.commands.RebootCommand,
):
    # Test reboot state management
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should manage reboot state appropriately
    if result == 0:
        output = capture.get()
        # Should handle state management
        assert any(msg in output for msg in ["Reboot initiated", "System will reboot"])
    else:
        assert result == 1



"""Cat command for Cascade."""

from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

if TYPE_CHECKING:
    import ops
    import shimmer

from rich.markdown import Markdown
from rich.syntax import Syntax

from ...utils.command_helpers import (
    create_file_progress,
    format_file_header,
    handle_help_flag,
    parse_flags,
    process_file_arguments,
    safe_read_file,
    validate_min_args,
)
from .._base import Command


class CatCommand(Command):
    """Display file contents."""

    name = "cat"
    help = "Display file contents: Usage: cat <file>"
    category = "Filesystem Commands"

    CODE_EXTENSIONS: ClassVar[dict[str, str]] = {
        ".py": "python",
        ".js": "javascript",
        ".ts": "typescript",
        ".sh": "bash",
        ".c": "c",
        ".cpp": "cpp",
        ".h": "c",
        ".hpp": "cpp",
        ".json": "json",
        ".yaml": "yaml",
        ".yml": "yaml",
        ".toml": "toml",
        ".md": "markdown",
        ".html": "html",
        ".css": "css",
        ".java": "java",
        ".go": "go",
        ".rs": "rust",
        ".rb": "ruby",
        ".pl": "perl",
        ".php": "php",
        ".xml": "xml",
    }

    def execute(
        self, client: ops.pebble.Client | shimmer.PebbleCliClient, args: list[str]
    ):
        """Execute cat command."""
        if handle_help_flag(self, args):
            return 0

        flags_result = parse_flags(args, {"plain": bool}, self.shell)
        if flags_result is None:
            return 1
        flags, remaining_args = flags_result

        if not validate_min_args(self.shell, remaining_args, 1, "cat <file>"):
            return 1

        file_paths = process_file_arguments(
            self.shell,
            client,
            remaining_args,
            allow_globs=True,
            min_files=1,
            max_files=1,
        )
        if file_paths is None:
            return 1

        # Process files with progress tracking
        with create_file_progress() as progress:
            task = progress.add_task("Reading files...", total=len(file_paths))

            for file_path in file_paths:
                # Read file content safely
                content = safe_read_file(client, file_path, self.shell)
                if content is None:
                    return 1

                # Print file header for multiple files
                header = format_file_header(file_path, len(file_paths))
                if header:
                    self.shell.console.print(header)

                # Display content based on file type and flags
                self._display_content(content, file_path, flags["plain"])
                progress.advance(task)

        return 0

    def _display_content(self, content: str, file_path: str, plain: bool) -> None:
        """Display file content with appropriate formatting."""
        if plain:
            self.shell.console.print(content, end="")
            return

        # Determine file extension for syntax highlighting
        ext: str | None = None
        for known_ext in self.CODE_EXTENSIONS:
            if file_path.endswith(known_ext):
                ext = known_ext
                break

        # Apply appropriate formatting
        if ext == ".md":
            md = Markdown(content, code_theme="monokai", justify="left")
            self.shell.console.print(md)
        elif ext == ".json":
            self.shell.console.print_json(content)
        elif ext and ext in self.CODE_EXTENSIONS:
            lexer = self.CODE_EXTENSIONS[ext]
            syntax = Syntax(
                content,
                lexer,
                theme="monokai",
                line_numbers=True,
                word_wrap=False,
            )
            self.shell.console.print(syntax)
        else:
            self.shell.console.print(content, end="")

        # Add newline between files if content doesn't end with one
        if content and not content.endswith("\n") and (plain or not ext):
            self.shell.console.print()



"""Integration tests for the ThemeCommand."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

import pebble_shell.commands
import pebble_shell.shell

if TYPE_CHECKING:
    import ops


@pytest.fixture
def command(shell: pebble_shell.shell.PebbleShell):
    """Fixture to create a ThemeCommand instance."""
    yield pebble_shell.commands.ThemeCommand(shell=shell)


def test_name(command: pebble_shell.commands.ThemeCommand):
    assert command.name == "theme"


def test_category(command: pebble_shell.commands.ThemeCommand):
    assert command.category == "System"


def test_help(command: pebble_shell.commands.ThemeCommand):
    with command.shell.console.capture() as capture:
        command.show_help()
    assert "Manage display themes" in capture.get()


@pytest.mark.parametrize("args", [["-h"], ["--help"]])
def test_execute_help(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ThemeCommand,
    args: list[str],
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=args)
    assert result == 0
    assert "Manage display themes" in capture.get()


def test_execute_no_args(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ThemeCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])
    assert result == 0
    assert "Usage: theme" in capture.get()


def test_execute_list(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ThemeCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["list"])
    assert result == 0
    output = capture.get()
    assert "Available themes:" in output
    assert "default" in output


def test_execute_show(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ThemeCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["show"])
    assert result == 0
    output = capture.get()
    assert "Current theme:" in output


def test_execute_set_default(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ThemeCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["set", "default"])
    assert result == 0
    output = capture.get()
    assert "Theme set to: default" in output


def test_execute_set_invalid(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ThemeCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["set", "nonexistent"])
    assert result == 1
    output = capture.get()
    assert "Unknown theme" in output


def test_execute_invalid_subcommand(
    client: ops.pebble.Client,
    command: pebble_shell.commands.ThemeCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["invalid"])
    assert result == 1
    output = capture.get()
    assert "Unknown subcommand" in output



"""Implementation of WhoCommand."""

from __future__ import annotations

from typing import TYPE_CHECKING

from rich.panel import Panel
from rich.text import Text

from ...utils.command_helpers import handle_help_flag
from ...utils.proc_reader import (
    ProcReadError,
    get_process_tty,
    get_user_name_for_uid,
    read_proc_cmdline,
    read_proc_status_field,
)
from ...utils.table_builder import create_standard_table
from .._base import Command

if TYPE_CHECKING:
    import ops
    import shimmer


class WhoCommand(Command):
    """Show logged in users."""

    name = "who"
    help = "Show logged in users"
    category = "System"

    def execute(
        self, client: ops.pebble.Client | shimmer.PebbleCliClient, args: list[str]
    ):
        """Execute who command with rich table output."""
        if handle_help_flag(self, args):
            return 0
        table = create_standard_table()
        table.add_column("USER", style="cyan", no_wrap=True)
        table.add_column("TTY", style="white", no_wrap=True)
        table.add_column("FROM", style="white", no_wrap=True)
        table.add_column("LOGIN@", style="white", no_wrap=True)
        table.add_column("IDLE", style="white", no_wrap=True)
        table.add_column("JCPU", style="white", no_wrap=True)
        table.add_column("PCPU", style="white", no_wrap=True)
        table.add_column("WHAT", style="green", no_wrap=False)

        proc_dirs: list[str] = []
        files = client.list_files("/proc")
        proc_dirs = [file_info.name for file_info in files if file_info.name.isdigit()]

        sessions: set[str] = set()
        for pid in proc_dirs:
            try:
                cmdline = read_proc_cmdline(client, pid)

                try:
                    uid = read_proc_status_field(client, pid, "Uid")
                except ProcReadError:
                    uid = None

                session_indicators = [
                    "ssh",
                    "sshd",
                    "login",
                    "bash",
                    "sh",
                    "zsh",
                    "fish",
                    "dash",
                    "ash",
                    "ksh",
                    "csh",
                    "tcsh",
                    "screen",
                    "tmux",
                ]
                if cmdline and any(
                    indicator in cmdline.lower() for indicator in session_indicators
                ):
                    username = (
                        get_user_name_for_uid(client, uid) or f"uid{uid}"
                        if uid
                        else "unknown"
                    )
                    tty = self._get_tty_info(client, pid)
                    session_id = f"{username}:{tty}:{cmdline[:20]}"
                    if session_id not in sessions:
                        sessions.add(session_id)
                        login_time = self._get_process_start_time(client, pid)
                        # For now, FROM, IDLE, JCPU, PCPU are not implemented (show as '-')
                        table.add_row(
                            f"[cyan]{username}[/cyan]",
                            tty,
                            "-",
                            login_time,
                            "-",
                            "-",
                            "-",
                            f"[green]{cmdline[:30]}[/green]",
                        )
            except Exception:  # noqa: S112, PERF203  # needed for process scanning
                continue

        if not sessions:
            self.console.print(
                Panel(
                    Text("No user sessions found", style="yellow"),
                    title="[b]who[/b]",
                    style="bold magenta",
                )
            )
            return 1
        self.console.print(table.build())
        return 0

    def _get_tty_info(
        self, client: ops.pebble.Client | shimmer.PebbleCliClient, pid: str
    ) -> str:
        """Get TTY information for a process."""
        try:
            return get_process_tty(client, pid)
        except ProcReadError:
            return "?"

    def _get_process_start_time(
        self, client: ops.pebble.Client | shimmer.PebbleCliClient, pid: str
    ) -> str:
        """Get process start time (simplified)."""
        with client.pull(f"/proc/{pid}/stat") as file:
            stat_content = file.read()
        assert isinstance(stat_content, str)
        # Parse stat file (starttime is the 22nd field).
        parts = stat_content.strip().split()
        if len(parts) >= 22:
            return parts[21]
        return "unknown"



"""Unit tests for bash-like quote handling in the parser."""

from pebble_shell.utils.parser import ShellParser, ShellVariables


class TestQuoteHandling:
    """Test bash-like quote handling functionality."""

    def setup_method(self):
        """Set up test fixtures."""
        self.variables = ShellVariables()
        self.variables.set_variable("TEST_VAR", "test_value")
        self.variables.set_variable("HOME", "/home/user")
        self.parser = ShellParser(self.variables)

    def test_single_quotes_literal(self):
        """Test that single quotes preserve everything literally."""
        result = self.parser._parse_with_bash_quotes("'hello world'")
        assert result == ["hello world"]

    def test_single_quotes_no_variable_expansion(self):
        """Test that variables are not expanded inside single quotes."""
        result = self.parser._parse_with_bash_quotes("'$TEST_VAR'")
        assert result == ["$TEST_VAR"]

    def test_double_quotes_preserve_spaces(self):
        """Test that double quotes preserve spaces."""
        result = self.parser._parse_with_bash_quotes('"hello world"')
        assert result == ["hello world"]

    def test_double_quotes_variable_expansion(self):
        """Test that variables are expanded inside double quotes."""
        result = self.parser._parse_with_bash_quotes('"$TEST_VAR"')
        assert result == ["test_value"]

    def test_double_quotes_braced_variable_expansion(self):
        """Test that ${VAR} variables are expanded inside double quotes."""
        result = self.parser._parse_with_bash_quotes('"${TEST_VAR}"')
        assert result == ["test_value"]

    def test_double_quotes_escape_sequences(self):
        """Test escape sequences inside double quotes."""
        result = self.parser._parse_with_bash_quotes(r'"hello \"world\""')
        assert result == ['hello "world"']

    def test_backslash_escaping_outside_quotes(self):
        """Test backslash escaping outside quotes."""
        result = self.parser._parse_with_bash_quotes(r"hello\ world")
        assert result == ["hello world"]

    def test_mixed_quoting(self):
        """Test mixing single and double quotes."""
        result = self.parser._parse_with_bash_quotes("'hello' \"world\"")
        assert result == ["hello", "world"]

    def test_variable_expansion_outside_quotes(self):
        """Test variable expansion works outside quotes."""
        result = self.parser._parse_with_bash_quotes("$TEST_VAR")
        assert result == ["test_value"]

    def test_braced_variable_expansion_outside_quotes(self):
        """Test ${VAR} variable expansion works outside quotes."""
        result = self.parser._parse_with_bash_quotes("${HOME}/file")
        assert result == ["/home/user/file"]

    def test_unknown_variable_expansion(self):
        """Test unknown variables expand to empty string."""
        result = self.parser._parse_with_bash_quotes("$UNKNOWN_VAR")
        assert result == [""]

    def test_special_variable_expansion(self):
        """Test special variables like $? expand correctly."""
        self.variables.last_exit_code = 42
        result = self.parser._parse_with_bash_quotes("$?")
        assert result == ["42"]

    def test_unclosed_single_quote(self):
        """Test handling of unclosed single quotes."""
        result = self.parser._parse_with_bash_quotes("'hello world")
        assert result == ["hello world"]  # Should consume to end

    def test_unclosed_double_quote(self):
        """Test handling of unclosed double quotes."""
        result = self.parser._parse_with_bash_quotes('"hello world')
        assert result == ["hello world"]  # Should consume to end

    def test_empty_quotes(self):
        """Test empty quoted strings."""
        result = self.parser._parse_with_bash_quotes('""')
        assert result == [""]

        result = self.parser._parse_with_bash_quotes("''")
        assert result == [""]

    def test_complex_quoting_scenario(self):
        """Test a complex scenario with mixed quoting and variables."""
        cmd = "\"$TEST_VAR\" 'literal $TEST_VAR' $HOME/file"
        result = self.parser._parse_with_bash_quotes(cmd)
        assert result == ["test_value", "literal $TEST_VAR", "/home/user/file"]

    def test_escaped_dollar_in_double_quotes(self):
        """Test escaped dollar sign in double quotes."""
        result = self.parser._parse_with_bash_quotes(r'"price is \$5"')
        assert result == ["price is $5"]

    def test_whitespace_handling(self):
        """Test proper whitespace handling with quotes."""
        result = self.parser._parse_with_bash_quotes('  "hello"   "world"  ')
        assert result == ["hello", "world"]

    def test_concatenated_quotes(self):
        """Test concatenated quoted strings."""
        result = self.parser._parse_with_bash_quotes('hello"world"test')
        assert result == ["helloworldtest"]

    def test_dollar_without_variable_name(self):
        """Test bare dollar sign handling."""
        result = self.parser._parse_with_bash_quotes('"$"')
        assert result == ["$"]  # Should remain as literal $

    def test_nested_brace_in_variable(self):
        """Test variable expansion with nested content."""
        self.variables.set_variable("PATH", "/bin:/usr/bin")
        result = self.parser._parse_with_bash_quotes('"${PATH}"')
        assert result == ["/bin:/usr/bin"]



"""Integration tests for the UnexpandCommand."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

import pebble_shell.commands
import pebble_shell.shell

if TYPE_CHECKING:
    import ops


@pytest.fixture
def command(shell: pebble_shell.shell.PebbleShell):
    """Fixture to create a UnexpandCommand instance."""
    yield pebble_shell.commands.UnexpandCommand(shell=shell)


def test_name(command: pebble_shell.commands.UnexpandCommand):
    assert command.name == "unexpand"


def test_category(command: pebble_shell.commands.UnexpandCommand):
    assert command.category == "Text Processing"


def test_help(command: pebble_shell.commands.UnexpandCommand):
    with command.shell.console.capture() as capture:
        command.show_help()
    output = capture.get()
    assert "unexpand" in output
    assert "Convert spaces to tabs" in output
    assert "spaces" in output


@pytest.mark.parametrize("args", [["-h"], ["--help"]])
def test_execute_help(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
    args: list[str],
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=args)
    assert result == 0
    output = capture.get()
    assert "unexpand" in output
    assert "Convert spaces to tabs" in output


def test_execute_no_args_stdin(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # unexpand with no args should read from stdin
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should either read from stdin or show usage
    if result == 0:
        output = capture.get()
        # Should wait for stdin input
        assert len(output) >= 0
    else:
        # Should fail with usage message
        assert result == 1
        output = capture.get()
        assert any(msg in output for msg in ["usage", "Usage", "unexpand", "file"])


def test_execute_convert_spaces_to_tabs_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test converting spaces to tabs in a file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    # Should either succeed converting or fail gracefully
    if result == 0:
        output = capture.get()
        # Should contain file contents with spaces converted to tabs
        assert len(output) >= 0
        # Should have converted leading spaces to tabs where appropriate
        if "    " in output:
            # Should contain tabs instead of multiple spaces
            assert "\t" in output
    else:
        # Should fail if file doesn't exist or access denied
        assert result == 1


def test_execute_all_spaces_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test converting all spaces, not just leading ones
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-a", "/etc/hosts"])

    # Should either succeed converting all spaces or fail gracefully
    if result == 0:
        output = capture.get()
        # Should convert all spaces to tabs, not just leading
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_first_only_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test converting only first sequence of spaces on each line
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["--first-only", "/etc/hosts"])

    # Should either succeed with first-only conversion or fail gracefully
    if result == 0:
        output = capture.get()
        # Should convert only first sequence of spaces
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_custom_tab_stops(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test custom tab stops
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-t", "4", "/etc/hosts"])

    # Should either succeed with custom tab stops or fail gracefully
    if result == 0:
        output = capture.get()
        # Should use 4-space tab stops
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_multiple_tab_stops(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test multiple custom tab stops
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-t", "4,8,12", "/etc/hosts"])

    # Should either succeed with multiple tab stops or fail gracefully
    if result == 0:
        output = capture.get()
        # Should use specified tab stops
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_leading_spaces_only(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test converting only leading spaces (default behavior)
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    # Should succeed converting leading spaces only
    if result == 0:
        output = capture.get()
        # Should convert leading spaces but leave internal spaces
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_preserve_existing_tabs(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test preserving existing tabs
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/fstab"])

    # Should either succeed preserving tabs or fail gracefully
    if result == 0:
        output = capture.get()
        # Should preserve existing tabs and add new ones
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_empty_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test with empty file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/dev/null"])

    # Should succeed with no output for empty file
    assert result == 0
    output = capture.get()
    # Should produce no output for empty file
    assert len(output) == 0


def test_execute_file_with_no_spaces(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test file with no spaces to convert
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/hostname"])

    # Should succeed with unchanged content
    if result == 0:
        output = capture.get()
        # Should pass through content unchanged if no spaces
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_file_with_only_tabs(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test file that already has only tabs
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/fstab"])

    # Should succeed preserving existing tabs
    if result == 0:
        output = capture.get()
        # Should preserve existing tab structure
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_mixed_spaces_and_tabs(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test file with mixed spaces and tabs
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    # Should succeed handling mixed whitespace
    if result == 0:
        output = capture.get()
        # Should handle mixed whitespace appropriately
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_lines_with_different_indentation(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test lines with varying indentation levels
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/hosts"])

    # Should succeed with varying indentation
    if result == 0:
        output = capture.get()
        # Should handle different indentation levels
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_very_long_lines(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test very long lines
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    # Should succeed with long lines
    if result == 0:
        output = capture.get()
        # Should handle long lines efficiently
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_unicode_content(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test file with Unicode content
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/hosts"])

    # Should either succeed with Unicode or fail gracefully
    if result == 0:
        output = capture.get()
        # Should handle Unicode characters properly
        assert len(output) >= 0
        # Should maintain Unicode character integrity
        if len(output) > 0:
            assert isinstance(output, str)
    else:
        assert result == 1


def test_execute_binary_file_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test with binary file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/bin/ls"])

    # Should either handle binary or fail gracefully
    if result == 0:
        output = capture.get()
        # Should process binary file (may contain binary data)
        assert len(output) >= 0
    else:
        # May fail for binary files
        assert result == 1


def test_execute_multiple_files(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test processing multiple files
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/hosts", "/etc/passwd"])

    # Should either succeed with multiple files or fail gracefully
    if result == 0:
        output = capture.get()
        # Should process both files
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_nonexistent_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test with nonexistent file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/nonexistent/file.txt"])

    # Should fail with file not found error
    assert result == 1
    output = capture.get()
    assert any(
        msg in output for msg in ["No such file", "cannot", "error", "not found"]
    )


def test_execute_permission_denied_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test with permission denied file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/root/.ssh/id_rsa"])

    # Should either succeed or fail with permission error
    if result == 1:
        output = capture.get()
        assert any(msg in output for msg in ["permission", "denied", "cannot", "error"])
    else:
        # May succeed if file is readable
        assert result == 0


def test_execute_directory_as_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test with directory instead of file
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc"])

    # Should fail with directory error
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["directory", "Is a directory", "error"])


def test_execute_invalid_tab_stops(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test with invalid tab stops
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-t", "abc", "/etc/hosts"])

    # Should fail with invalid tab stops error
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["invalid", "tab", "number", "error"])


def test_execute_zero_tab_stops(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test with zero tab stops
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-t", "0", "/etc/hosts"])

    # Should either handle zero or fail appropriately
    if result == 1:
        output = capture.get()
        assert any(msg in output for msg in ["invalid", "zero", "tab", "error"])
    else:
        # May handle zero tab stops
        assert result == 0


def test_execute_negative_tab_stops(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test with negative tab stops
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-t", "-4", "/etc/hosts"])

    # Should fail with negative tab stops error
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["invalid", "negative", "tab", "error"])


def test_execute_very_large_tab_stops(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test with very large tab stops
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-t", "999999", "/etc/hosts"])

    # Should either handle large tab stops or fail appropriately
    if result == 0:
        output = capture.get()
        # Should handle large tab stops
        assert len(output) >= 0
    else:
        # Should fail if tab stops too large
        assert result == 1


def test_execute_conflicting_options(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test conflicting options
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["-a", "--first-only", "/etc/hosts"]
        )

    # Should either handle conflicts or fail appropriately
    if result == 1:
        output = capture.get()
        assert any(msg in output for msg in ["conflicting", "invalid", "error"])
    else:
        # May resolve conflicts automatically
        assert result == 0


def test_execute_invalid_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test invalid option
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-z", "/etc/hosts"])

    # Should fail with invalid option error
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["invalid", "unknown", "option", "usage"])


def test_execute_output_format_validation(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test output format validation
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/hosts"])

    # Should produce properly formatted output
    if result == 0:
        output = capture.get()
        # Should maintain text structure
        assert len(output) >= 0
        if len(output) > 0:
            # Should preserve line structure
            lines = output.split("\n")
            assert len(lines) >= 1
    else:
        assert result == 1


def test_execute_tab_alignment_correctness(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test tab alignment correctness
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-t", "8", "/etc/passwd"])

    # Should produce correctly aligned tabs
    if result == 0:
        output = capture.get()
        # Should align to 8-column tab stops
        assert len(output) >= 0
        # Tabs should align content properly
        if "\t" in output:
            assert len(output) > 0
    else:
        assert result == 1


def test_execute_space_to_tab_conversion_accuracy(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test accuracy of space to tab conversion
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/hosts"])

    # Should accurately convert spaces to tabs
    if result == 0:
        output = capture.get()
        # Should convert appropriate spaces to tabs
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_memory_efficiency(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test memory efficiency
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    # Should be memory efficient
    if result == 0:
        output = capture.get()
        # Should not consume excessive memory
        assert len(output) < 100000  # Reasonable output size limit
    else:
        assert result == 1


def test_execute_concurrent_execution(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test concurrent execution safety
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/hosts"])

    # Should handle concurrent execution safely
    if result == 0:
        output = capture.get()
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_error_recovery(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test error recovery capabilities
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-z", "/etc/hosts"])

    # Should recover from errors gracefully
    assert result == 1
    output = capture.get()
    # Should provide meaningful error message
    assert any(msg in output for msg in ["invalid", "unknown", "error", "usage"])


def test_execute_signal_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test signal handling during processing
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/hosts"])

    # Should handle signals appropriately
    if result == 0:
        output = capture.get()
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_locale_independence(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test locale independence
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/hosts"])

    # Should work regardless of locale settings
    if result == 0:
        output = capture.get()
        # Should be locale-independent
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_cross_platform_compatibility(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test cross-platform compatibility
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/hosts"])

    # Should work across different platforms
    if result == 0:
        output = capture.get()
        # Should handle platform-specific features
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_performance_optimization(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test performance optimization
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    # Should complete efficiently
    if result == 0:
        output = capture.get()
        # Should process efficiently
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_data_consistency(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test data consistency
    with command.shell.console.capture() as _:
        result1 = command.execute(client=client, args=["/etc/hosts"])

    with command.shell.console.capture() as _:
        result2 = command.execute(client=client, args=["/etc/hosts"])

    # Should produce consistent results
    if result1 == 0 and result2 == 0:
        # Both executions should succeed consistently
        assert result1 == result2
    else:
        # At least one should succeed or both should fail consistently
        assert result1 == result2


def test_execute_robust_operation(
    client: ops.pebble.Client,
    command: pebble_shell.commands.UnexpandCommand,
):
    # Test robust operation under stress
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    # Should operate robustly
    if result == 0:
        output = capture.get()
        # Should handle stress conditions
        assert len(output) >= 0
    else:
        assert result == 1



"""Integration tests for the LsmodCommand."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

import pebble_shell.commands
import pebble_shell.shell

if TYPE_CHECKING:
    import ops


@pytest.fixture
def command(shell: pebble_shell.shell.PebbleShell):
    """Fixture to create a LsmodCommand instance."""
    yield pebble_shell.commands.LsmodCommand(shell=shell)


def test_name(command: pebble_shell.commands.LsmodCommand):
    assert command.name == "lsmod"


def test_category(command: pebble_shell.commands.LsmodCommand):
    assert command.category == "System Information"


def test_help(command: pebble_shell.commands.LsmodCommand):
    with command.shell.console.capture() as capture:
        command.show_help()
    output = capture.get()
    assert "lsmod" in output
    assert any(
        phrase in output.lower() for phrase in ["module", "kernel", "list", "loaded"]
    )


@pytest.mark.parametrize("args", [["-h"], ["--help"]])
def test_execute_help(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LsmodCommand,
    args: list[str],
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=args)
    assert result == 0
    output = capture.get()
    assert "lsmod" in output
    assert any(
        phrase in output.lower() for phrase in ["module", "kernel", "list", "loaded"]
    )


def test_execute_no_args_list_modules(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LsmodCommand,
):
    # lsmod with no args should list all loaded modules
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should either succeed listing modules or fail gracefully
    if result == 0:
        output = capture.get()
        # Should contain module information
        assert len(output) >= 0
        if len(output) > 0:
            # Should contain module header or module names
            lines = output.split("\n")
            assert len(lines) >= 1
            # First line might be header: "Module Size Used by"
            if len(lines) > 1:
                header = lines[0].lower()
                assert any(word in header for word in ["module", "size", "used"])
    else:
        # Should fail if cannot access module information
        assert result == 1


def test_execute_basic_module_listing(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LsmodCommand,
):
    # Test basic module listing functionality
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should either succeed showing modules or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show loaded kernel modules
        assert len(output) >= 0
        if len(output) > 0:
            # Should have tabular format
            lines = output.split("\n")
            for line in lines:
                if line.strip() and not line.startswith("Module"):
                    # Each module line should have multiple columns
                    columns = line.split()
                    if len(columns) >= 2:
                        # Should have module name and size
                        assert len(columns[0]) > 0  # Module name
                        assert columns[1].isdigit() or columns[1] == "0"  # Size
    else:
        assert result == 1


def test_execute_module_format_validation(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LsmodCommand,
):
    # Test module output format validation
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should either succeed with proper format or fail gracefully
    if result == 0:
        output = capture.get()
        # Should have proper module format
        assert len(output) >= 0
        if len(output) > 0:
            lines = output.split("\n")
            non_empty_lines = [line for line in lines if line.strip()]
            if len(non_empty_lines) > 0:
                # Should have consistent column structure
                first_line = non_empty_lines[0]
                if "Module" in first_line:
                    # Header line format
                    assert any(
                        word in first_line for word in ["Module", "Size", "Used"]
                    )
    else:
        assert result == 1


def test_execute_kernel_module_detection(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LsmodCommand,
):
    # Test kernel module detection
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should either succeed detecting modules or fail gracefully
    if result == 0:
        output = capture.get()
        # Should detect kernel modules
        assert len(output) >= 0
        # May have no modules loaded (valid case)
        if len(output) > 0:
            lines = output.split("\n")
            # Should show some module information
            assert len(lines) >= 1
    else:
        assert result == 1


def test_execute_module_size_reporting(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LsmodCommand,
):
    # Test module size reporting
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should either succeed reporting sizes or fail gracefully
    if result == 0:
        output = capture.get()
        # Should report module sizes
        assert len(output) >= 0
        if len(output) > 0:
            lines = output.split("\n")
            for line in lines:
                if line.strip() and not line.startswith("Module"):
                    columns = line.split()
                    if len(columns) >= 2:
                        # Size should be numeric
                        size = columns[1]
                        assert size.isdigit() or size == "0"
    else:
        assert result == 1


def test_execute_module_usage_reporting(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LsmodCommand,
):
    # Test module usage count reporting
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should either succeed reporting usage or fail gracefully
    if result == 0:
        output = capture.get()
        # Should report module usage counts
        assert len(output) >= 0
        if len(output) > 0:
            lines = output.split("\n")
            for line in lines:
                if line.strip() and not line.startswith("Module"):
                    columns = line.split()
                    if len(columns) >= 3:
                        # Usage count should be numeric
                        usage = columns[2]
                        assert usage.isdigit() or usage == "0" or usage == "-"
    else:
        assert result == 1


def test_execute_module_dependencies(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LsmodCommand,
):
    # Test module dependency reporting
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should either succeed showing dependencies or fail gracefully
    if result == 0:
        output = capture.get()
        # Should show module dependencies
        assert len(output) >= 0
        if len(output) > 0:
            lines = output.split("\n")
            for line in lines:
                if line.strip() and not line.startswith("Module"):
                    # Dependencies shown in fourth column
                    columns = line.split(None, 3)  # Split into max 4 parts
                    if len(columns) >= 4:
                        # Dependencies column exists
                        deps = columns[3]
                        # Dependencies can be module names or empty
                        assert isinstance(deps, str)
    else:
        assert result == 1


def test_execute_proc_modules_access(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LsmodCommand,
):
    # Test access to /proc/modules
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should either succeed accessing proc or fail gracefully
    if result == 0:
        output = capture.get()
        # Should access /proc/modules successfully
        assert len(output) >= 0
    else:
        # Should fail if /proc/modules not accessible
        assert result == 1
        output = capture.get()
        assert any(
            msg in output for msg in ["cannot", "access", "proc", "permission", "error"]
        )


def test_execute_empty_module_list(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LsmodCommand,
):
    # Test with empty module list (no modules loaded)
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should either succeed with empty list or fail gracefully
    if result == 0:
        output = capture.get()
        # Should handle empty module list
        assert len(output) >= 0
        if len(output) == 0 or output.strip() == "":
            # No modules loaded - valid case
            assert True
        else:
            # May show header only
            lines = output.split("\n")
            non_empty = [line for line in lines if line.strip()]
            if len(non_empty) == 1:
                # Just header line
                assert "Module" in non_empty[0]
    else:
        assert result == 1


def test_execute_invalid_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LsmodCommand,
):
    # Test invalid option
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-z"])

    # Should fail with invalid option error
    assert result == 1
    output = capture.get()
    assert any(msg in output for msg in ["invalid", "unknown", "option", "usage"])


def test_execute_unexpected_arguments(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LsmodCommand,
):
    # Test with unexpected arguments
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["unexpected", "args"])

    # Should either ignore args or fail gracefully
    if result == 0:
        output = capture.get()
        # Should ignore unexpected arguments
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_permission_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LsmodCommand,
):
    # Test permission handling
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should either succeed with permissions or fail gracefully
    if result == 0:
        output = capture.get()
        # Should access module information
        assert len(output) >= 0
    else:
        # Should fail if insufficient permissions
        assert result == 1
        output = capture.get()
        assert any(msg in output for msg in ["permission", "denied", "access", "error"])


def test_execute_system_integration(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LsmodCommand,
):
    # Test system integration
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should integrate with kernel module system
    if result == 0:
        output = capture.get()
        # Should integrate properly
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_output_consistency(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LsmodCommand,
):
    # Test output consistency across multiple calls
    with command.shell.console.capture() as _:
        result1 = command.execute(client=client, args=[])

    with command.shell.console.capture() as _:
        result2 = command.execute(client=client, args=[])

    # Should produce consistent results
    if result1 == 0 and result2 == 0:
        # Both should succeed with similar output structure
        assert result1 == result2
    else:
        # Should fail consistently
        assert result1 == result2


def test_execute_large_module_list(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LsmodCommand,
):
    # Test with large number of modules
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should handle large module lists efficiently
    if result == 0:
        output = capture.get()
        # Should handle many modules
        assert len(output) >= 0
        if len(output) > 0:
            lines = output.split("\n")
            # Should handle numerous modules without issues
            assert len(lines) >= 1
    else:
        assert result == 1


def test_execute_memory_efficiency(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LsmodCommand,
):
    # Test memory efficiency
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should be memory efficient
    if result == 0:
        output = capture.get()
        # Should not consume excessive memory
        assert len(output) < 100000  # Reasonable output size limit
    else:
        assert result == 1


def test_execute_performance_optimization(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LsmodCommand,
):
    # Test performance optimization
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should complete efficiently
    if result == 0:
        output = capture.get()
        # Should process efficiently
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_error_recovery(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LsmodCommand,
):
    # Test error recovery capabilities
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["--invalid"])

    # Should recover from errors gracefully
    assert result == 1
    output = capture.get()
    # Should provide meaningful error message
    assert any(msg in output for msg in ["invalid", "unknown", "error", "usage"])


def test_execute_signal_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LsmodCommand,
):
    # Test signal handling during module listing
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should handle signals appropriately
    if result == 0:
        output = capture.get()
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_concurrent_execution(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LsmodCommand,
):
    # Test concurrent execution safety
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should handle concurrent execution safely
    if result == 0:
        output = capture.get()
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_cross_platform_compatibility(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LsmodCommand,
):
    # Test cross-platform compatibility
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should work across different platforms
    if result == 0:
        output = capture.get()
        # Should handle platform differences
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_kernel_version_compatibility(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LsmodCommand,
):
    # Test kernel version compatibility
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should work with different kernel versions
    if result == 0:
        output = capture.get()
        # Should be compatible across kernel versions
        assert len(output) >= 0
    else:
        assert result == 1


def test_execute_module_state_consistency(
    client: ops.pebble.Client,
    command: pebble_shell.commands.LsmodCommand,
):
    # Test module state consistency
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])

    # Should show consistent module state
    if result == 0:
        output = capture.get()
        # Should reflect current module state
        assert len(output) >= 0
    else:
        assert result == 1



"""Implementation of HdCommand."""

from __future__ import annotations

from typing import TYPE_CHECKING, Union

import ops

from .hexdump import HexdumpCommand

if TYPE_CHECKING:
    import shimmer

# TODO: Use the prototype from Shimmer.
ClientType = Union[ops.pebble.Client, "shimmer.PebbleCliClient"]


class HdCommand(HexdumpCommand):
    """Implementation of hd command (alias for hexdump -C)."""

    name = "hd"
    help = "Display file contents in hexadecimal (canonical format)"
    category = "Built-in Commands"

    def execute(
        self, client: ops.pebble.Client | shimmer.PebbleCliClient, args: list[str]
    ) -> int:
        """Execute the hd command."""
        # hd is equivalent to hexdump -C, so prepend -C to args
        return super().execute(client, ["-C", *args])



"""Implementation of SyslogCommand."""

from __future__ import annotations

import re
import time
from typing import TYPE_CHECKING

import ops
from rich.live import Live
from rich.panel import Panel
from rich.text import Text

from ...utils.command_helpers import handle_help_flag
from .._base import Command

if TYPE_CHECKING:
    import shimmer


# TODO: The syslog can be very large, and Pebble can only get the whole file.
# Can we do anything about that?


class SyslogCommand(Command):
    """Show syslog information from /var/log/syslog or similar files."""

    name = "syslog"
    help = "Show syslog information. Use -n NUM for last NUM lines, -f to follow, and an optional pattern to filter."
    category = "System"

    def execute(
        self, client: ops.pebble.Client | shimmer.PebbleCliClient, args: list[str]
    ):
        """Execute syslog command."""
        if handle_help_flag(self, args):
            return 0
        num_lines = 100
        follow = False
        pattern = None
        i = 0
        while i < len(args):
            arg = args[i]
            if arg == "-n" and i + 1 < len(args):
                try:
                    num_lines = int(args[i + 1])
                except ValueError:
                    self.console.print("[red]Invalid number for -n[/red]")
                    return 1
                i += 1
            elif arg == "-f":
                follow = True
            elif not arg.startswith("-") and pattern is None:
                pattern = arg
            i += 1

        log_files = [
            "/var/log/syslog",
            "/var/log/messages",
            "/var/log/system.log",
            "/var/log/user.log",
        ]
        log_file = None
        content = ""
        for lf in log_files:
            try:
                with client.pull(lf) as file:
                    content = file.read()
                assert isinstance(content, str)
            except ops.pebble.PathError:
                continue
            log_file = lf
            break
        else:
            self.console.print(
                Panel(
                    Text(
                        "No syslog file found. Tried: " + ", ".join(log_files),
                        style="yellow",
                    ),
                    title="[b]syslog[/b]",
                    style="bold magenta",
                )
            )
            return 1
        lines = content.splitlines()
        if pattern:
            regex = re.compile(pattern, re.IGNORECASE)
            lines = [line for line in lines if regex.search(line)]
        if num_lines > 0:
            lines = lines[-num_lines:]
        self._print_syslog_lines(lines)
        if follow:
            self._follow_syslog(client, log_file, pattern, num_lines)
        return 0

    def _print_syslog_lines(self, lines: list[str]):
        for line in lines:
            style = self._get_line_style(line)
            self.console.print(Text(line, style=style))

    def _get_line_style(self, line: str) -> str:
        if any(word in line.lower() for word in ["error", "fail", "critical"]):
            return "bold red"
        if any(word in line.lower() for word in ["warn", "deprecated"]):
            return "yellow"
        if any(word in line.lower() for word in ["info", "started", "listening"]):
            return "green"
        if any(word in line.lower() for word in ["debug"]):
            return "cyan"
        return "white"

    def _follow_syslog(
        self,
        client: ops.pebble.Client | shimmer.PebbleCliClient,
        log_file: str,
        pattern: str | None,
        num_lines: int,
    ):
        regex = re.compile(pattern, re.IGNORECASE) if pattern else None
        with Live(refresh_per_second=2, screen=False) as live:
            pos = 0
            while True:
                try:
                    with client.pull(log_file) as file:
                        content = file.read()
                    assert isinstance(content, str)
                    lines = content.splitlines()
                    if regex:
                        lines = [line for line in lines if regex.search(line)]
                    if num_lines > 0:
                        lines = lines[-num_lines:]
                    new_lines = lines[pos:]
                    if new_lines:
                        for line in new_lines:
                            style = self._get_line_style(line)
                            live.console.print(Text(line, style=style))
                        pos = len(lines)
                    time.sleep(1)
                except KeyboardInterrupt:  # noqa: PERF203
                    break



"""Implementation of LastCommand."""

from __future__ import annotations

from typing import TYPE_CHECKING

import ops
from rich.panel import Panel
from rich.text import Text

from ...utils.command_helpers import handle_help_flag
from ...utils.table_builder import create_enhanced_table
from .._base import Command

if TYPE_CHECKING:
    import shimmer


class LastCommand(Command):
    """Show last login information."""

    name = "last"
    help = "Show last login information"
    category = "System"

    def execute(
        self, client: ops.pebble.Client | shimmer.PebbleCliClient, args: list[str]
    ):
        """Execute last command."""
        if handle_help_flag(self, args):
            return 0
        # Try different log files for login information
        log_files = [
            "/var/log/auth.log",
            "/var/log/secure",
            "/var/log/wtmp",
            "/var/log/btmp",
        ]

        login_entries: list[tuple[str, str]] = []

        for log_file in log_files:
            try:
                with client.pull(log_file) as file:
                    content = file.read()
                assert isinstance(content, str)
            except ops.pebble.PathError:
                continue

            # Use list.extend for better performance
            login_entries.extend(
                (log_file, line)
                for line in content.splitlines()
                if any(
                    keyword in line.lower()
                    for keyword in [
                        "session opened",
                        "accepted",
                        "login",
                        "sshd",
                    ]
                )
            )

        if not login_entries:
            self.console.print(
                Panel(
                    Text("No login information found", style="yellow"),
                    title="[b]last[/b]",
                    style="bold magenta",
                )
            )
            return 1

        # Sort by most recent (simplified - just take last 50 entries)
        login_entries = login_entries[-50:]

        table = create_enhanced_table()
        table.add_column("User", style="cyan", no_wrap=True)
        table.add_column("TTY", style="green", no_wrap=True)
        table.add_column("From", style="yellow", no_wrap=True)
        table.add_column("Login Time", style="blue", no_wrap=True)
        table.add_column("Logout Time", style="magenta", no_wrap=True)
        table.add_column("Duration", style="white", no_wrap=True)

        for _log_file, line in login_entries:
            # Parse login line (simplified parsing)
            parts = line.split()
            if len(parts) >= 5:
                # Try to extract user, tty, from, and time
                user = "unknown"
                tty = "?"
                from_host = "?"
                login_time = "unknown"

                # Look for common patterns
                for i, part in enumerate(parts):
                    if part in ["session", "opened", "accepted", "login"]:
                        if i > 0:
                            user = parts[i - 1]
                        break

                # Extract timestamp (usually at the beginning)
                if len(parts) >= 3:
                    login_time = " ".join(parts[:3])

                table.add_row(user, tty, from_host, login_time, "-", "-")

        self.console.print(table.build())
        return 0



"""Implementation of SumCommand."""

from __future__ import annotations

from typing import TYPE_CHECKING, Union

import ops

from ...utils.command_helpers import handle_help_flag, parse_flags, safe_read_file
from ...utils.theme import get_theme
from .._base import Command

if TYPE_CHECKING:
    import shimmer

# TODO: Use the prototype from Shimmer.
ClientType = Union[ops.pebble.Client, "shimmer.PebbleCliClient"]


class SumCommand(Command):
    """Implementation of sum command."""

    name = "sum"
    help = "Calculate and display checksums"
    category = "File Utilities"

    def show_help(self):
        """Show command help."""
        help_text = """Calculate and display checksums.

Usage: sum [OPTIONS] [FILE...]

Description:
    Calculate and display BSD-style or System V-style checksums.
    If no files are specified, read from standard input.

Options:
    -r              Use BSD sum algorithm (default)
    -s, --sysv      Use System V sum algorithm
    -h, --help      Show this help message

Examples:
    sum file.txt
    sum -s file1.txt file2.txt
        """
        self.console.print(help_text)

    def execute(self, client: ClientType, args: list[str]) -> int:
        """Execute the sum command."""
        if handle_help_flag(self, args):
            return 0

        parse_result = parse_flags(
            args,
            {
                "r": bool,  # BSD algorithm (default)
                "s": bool,  # System V algorithm
                "sysv": bool,  # System V algorithm
            },
            self.shell,
        )
        if parse_result is None:
            return 1
        flags, positional_args = parse_result

        use_sysv = flags.get("s", False) or flags.get("sysv", False)
        files = positional_args if positional_args else ["-"]
        exit_code = 0

        for file_path in files:
            try:
                if file_path == "-":
                    # TODO: Handle stdin input
                    self.console.print(
                        get_theme().warning_text(
                            "sum: reading from stdin not supported"
                        )
                    )
                    continue

                # Read file from remote system
                data = safe_read_file(client, file_path, self.shell)
                if data is None:
                    self.console.print(
                        get_theme().error_text(
                            f"sum: {file_path}: No such file or directory"
                        )
                    )
                    exit_code = 1
                    continue

                # Convert string to bytes for checksum calculation
                data_bytes = data.encode("utf-8")

                if use_sysv:
                    # System V algorithm: sum of all bytes
                    checksum = sum(data_bytes) & 0xFFFF
                    blocks = (len(data_bytes) + 511) // 512  # 512-byte blocks
                    self.console.print(f"{checksum} {blocks} {file_path}")
                else:
                    # BSD algorithm: rotating checksum
                    checksum = 0
                    for byte in data_bytes:
                        checksum = ((checksum >> 1) | ((checksum & 1) << 15)) + byte
                        checksum &= 0xFFFF

                    blocks = (len(data_bytes) + 1023) // 1024  # 1024-byte blocks
                    self.console.print(f"{checksum:05d} {blocks:5d} {file_path}")

            except Exception as e:
                self.console.print(get_theme().error_text(f"sum: {file_path}: {e}"))
                exit_code = 1

        return exit_code



"""Implementation of InfoCommand."""

from __future__ import annotations

from typing import TYPE_CHECKING, Union

import ops

from ...utils.command_helpers import handle_help_flag
from .._base import Command

if TYPE_CHECKING:
    import shimmer

# TODO: Use the prototype from Shimmer.
ClientType = Union[ops.pebble.Client, "shimmer.PebbleCliClient"]


class InfoCommand(Command):
    """Command for displaying system and shell information."""

    name = "info"
    help = "Show system information"
    category = "Built-in Commands"

    def execute(
        self, client: ops.pebble.Client | shimmer.PebbleCliClient, args: list[str]
    ):
        """Execute the info command to display system information."""
        if handle_help_flag(self, args):
            return 0
        self.console.print("=== Cascade System Information ===")
        self.console.print()

        self.console.print("Shell Information:")
        self.console.print(f"  Current Directory: {self.shell.current_directory}")
        self.console.print()

        # TODO: We should add more things here!

        self.console.print("Use 'help' to see all shell features")
        return 0



"""Integration tests for the TeeCommand."""

from __future__ import annotations

import tempfile
from typing import TYPE_CHECKING

import pytest

import pebble_shell.commands
import pebble_shell.shell

if TYPE_CHECKING:
    import ops


@pytest.fixture
def command(shell: pebble_shell.shell.PebbleShell):
    """Fixture to create a TeeCommand instance."""
    yield pebble_shell.commands.TeeCommand(shell=shell)


def test_name(command: pebble_shell.commands.TeeCommand):
    assert command.name == "tee"


def test_category(command: pebble_shell.commands.TeeCommand):
    assert command.category == "Advanced Utilities"


def test_help(command: pebble_shell.commands.TeeCommand):
    with command.shell.console.capture() as capture:
        command.show_help()
    assert "copy" in capture.get() or "write" in capture.get()


@pytest.mark.parametrize("args", [["-h"], ["--help"]])
def test_execute_help(
    client: ops.pebble.Client,
    command: pebble_shell.commands.TeeCommand,
    args: list[str],
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=args)
    assert result == 0
    assert "tee" in capture.get() or "copy" in capture.get()


def test_execute_no_args(
    client: ops.pebble.Client,
    command: pebble_shell.commands.TeeCommand,
):
    with command.shell.console.capture() as _:
        result = command.execute(client=client, args=[])
    # tee without files should succeed (read from stdin, write to stdout)
    assert result == 0


def test_execute_with_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.TeeCommand,
):
    with tempfile.NamedTemporaryFile(suffix="_test_tee.txt", delete=False) as temp_file:
        temp_path = temp_file.name
    with command.shell.console.capture() as _:
        result = command.execute(client=client, args=[temp_path])
    # Should succeed writing to temporary file
    assert result == 0


def test_execute_with_append_flag(
    client: ops.pebble.Client,
    command: pebble_shell.commands.TeeCommand,
):
    with tempfile.NamedTemporaryFile(suffix="_test_tee.txt", delete=False) as temp_file:
        temp_path = temp_file.name
    with command.shell.console.capture() as _:
        result = command.execute(client=client, args=["-a", temp_path])
    # Should succeed with append flag to temporary file
    assert result == 0


def test_execute_with_ignore_interrupts_flag(
    client: ops.pebble.Client,
    command: pebble_shell.commands.TeeCommand,
):
    with tempfile.NamedTemporaryFile(suffix="_test_tee.txt", delete=False) as temp_file:
        temp_path = temp_file.name
    with command.shell.console.capture() as _:
        result = command.execute(client=client, args=["-i", temp_path])
    # Should succeed with ignore interrupts flag to temporary file
    assert result == 0


def test_execute_with_multiple_files(
    client: ops.pebble.Client,
    command: pebble_shell.commands.TeeCommand,
):
    with tempfile.NamedTemporaryFile(suffix="_file1.txt", delete=False) as temp_file1:
        temp_path1 = temp_file1.name
    with tempfile.NamedTemporaryFile(suffix="_file2.txt", delete=False) as temp_file2:
        temp_path2 = temp_file2.name
    with command.shell.console.capture() as _:
        result = command.execute(client=client, args=[temp_path1, temp_path2])
    # Should succeed writing to multiple temporary files
    assert result == 0


def test_execute_with_combined_flags(
    client: ops.pebble.Client,
    command: pebble_shell.commands.TeeCommand,
):
    with tempfile.NamedTemporaryFile(suffix="_test_tee.txt", delete=False) as temp_file:
        temp_path = temp_file.name
    with command.shell.console.capture() as _:
        result = command.execute(client=client, args=["-ai", temp_path])
    # Should succeed with combined flags to temporary file
    assert result == 0



"""Split command for Cascade."""

from __future__ import annotations

from typing import TYPE_CHECKING, Union

import ops

from ...utils.command_helpers import handle_help_flag, parse_flags, safe_read_file
from ...utils.theme import get_theme
from .._base import Command

if TYPE_CHECKING:
    import shimmer


# TODO: Use the prototype from Shimmer.
ClientType = Union[ops.pebble.Client, "shimmer.PebbleCliClient"]


class SplitCommand(Command):
    """Implementation of split command."""

    name = "split"
    help = "Split a file into pieces"
    category = "Data Processing"

    def show_help(self):
        """Show command help."""
        help_text = """Split a file into pieces.

Usage: split [OPTIONS] [INPUT [PREFIX]]

Description:
    Output fixed-size pieces of INPUT to PREFIXaa, PREFIXab, ...
    Default PREFIX is 'x'.

Options:
    -a SUFFIX_LENGTH Use SUFFIX_LENGTH characters for suffix (default: 2)
    -b SIZE         Put SIZE bytes per output file
    -C SIZE         Put at most SIZE bytes per line per output file
    -l NUMBER       Put NUMBER lines per output file (default: 1000)
    -d              Use numeric suffixes instead of alphabetic
    --verbose       Print a diagnostic just before each output file is opened
    -h, --help      Show this help message

SIZE can be:
    N               N bytes
    Nk or NK        N kilobytes
    Nm or NM        N megabytes
    Ng or NG        N gigabytes

Examples:
    split file.txt
    split -l 500 file.txt part_
    split -b 1024 file.txt chunk_
        """
        self.console.print(help_text)

    def execute(self, client: ClientType, args: list[str]) -> int:
        """Execute the split command."""
        if handle_help_flag(self, args):
            return 0

        # TODO: Can this use common flag parsing code?
        parse_result = parse_flags(
            args,
            {
                "a": str,  # suffix length
                "b": str,  # bytes per file
                "C": str,  # bytes per line per file
                "l": str,  # lines per file
                "d": bool,  # numeric suffixes
                "verbose": bool,  # verbose output
            },
            self.shell,
        )
        if parse_result is None:
            return 1
        flags, positional_args = parse_result

        suffix_length = int(flags.get("a") or 2)
        bytes_per_file = self._parse_size(flags.get("b"))
        bytes_per_line = self._parse_size(flags.get("C"))
        lines_per_file = int(flags.get("l") or 1000)
        numeric_suffix = flags.get("d", False)
        verbose = flags.get("verbose", False)

        # Get input file and prefix
        input_file = positional_args[0] if positional_args else "-"
        prefix = positional_args[1] if len(positional_args) > 1 else "x"

        # Determine split mode
        if bytes_per_file:
            split_mode = "bytes"
            split_size = bytes_per_file
        elif bytes_per_line:
            split_mode = "line_bytes"
            split_size = bytes_per_line
        else:
            split_mode = "lines"
            split_size = lines_per_file

        try:
            # Read input
            if input_file == "-":
                # TODO: Handle stdin
                self.console.print(
                    get_theme().warning_text("split: reading from stdin not supported")
                )
                return 1
            else:
                content = safe_read_file(client, input_file, self.shell)
                if content is None:
                    return 1

            # Split the content
            if split_mode == "lines":
                self._split_by_lines(
                    client,
                    content,
                    prefix,
                    split_size,
                    suffix_length,
                    numeric_suffix,
                    verbose,
                )
            elif split_mode == "bytes":
                self._split_by_bytes(
                    client,
                    content,
                    prefix,
                    split_size,
                    suffix_length,
                    numeric_suffix,
                    verbose,
                )
            elif split_mode == "line_bytes":
                self._split_by_line_bytes(
                    client,
                    content,
                    prefix,
                    split_size,
                    suffix_length,
                    numeric_suffix,
                    verbose,
                )

            return 0

        except ops.pebble.PathError:
            self.console.print(
                get_theme().error_text(
                    f"split: {input_file}: No such file or directory"
                )
            )
            return 1
        except Exception as e:
            self.console.print(get_theme().error_text(f"split: {e}"))
            return 1

    def _parse_size(self, size_str: str | None) -> int | None:
        """Parse size string (e.g., '1024', '1K', '1M')."""
        if not size_str:
            return None

        size_str = size_str.upper()
        multipliers = {
            "K": 1024,
            "M": 1024 * 1024,
            "G": 1024 * 1024 * 1024,
        }

        if size_str[-1] in multipliers:
            return int(size_str[:-1]) * multipliers[size_str[-1]]
        else:
            return int(size_str)

    def _generate_suffix(self, index: int, length: int, numeric: bool) -> str:
        """Generate suffix for split files."""
        if numeric:
            return f"{index:0{length}d}"
        else:
            # Alphabetic suffix (aa, ab, ac, ..., ba, bb, ...)
            suffix = ""
            remaining = index
            for _ in range(length):
                suffix = chr(ord("a") + (remaining % 26)) + suffix
                remaining //= 26
            return suffix

    def _split_by_lines(
        self,
        client: ClientType,
        content: str,
        prefix: str,
        lines_per_file: int,
        suffix_length: int,
        numeric_suffix: bool,
        verbose: bool,
    ):
        """Split content by number of lines."""
        lines = content.splitlines(keepends=True)

        for file_index, i in enumerate(range(0, len(lines), lines_per_file)):
            chunk_lines = lines[i : i + lines_per_file]
            chunk_content = "".join(chunk_lines)

            suffix = self._generate_suffix(file_index, suffix_length, numeric_suffix)
            output_filename = f"{prefix}{suffix}"

            if verbose:
                self.console.print(f"creating file '{output_filename}'")

            # Write to remote file
            from io import StringIO

            output_stream = StringIO(chunk_content)
            client.push(output_filename, output_stream)

    def _split_by_bytes(
        self,
        client: ClientType,
        content: str,
        prefix: str,
        bytes_per_file: int,
        suffix_length: int,
        numeric_suffix: bool,
        verbose: bool,
    ):
        """Split content by number of bytes."""
        content_bytes = content.encode("utf-8")

        for file_index, i in enumerate(range(0, len(content_bytes), bytes_per_file)):
            chunk_bytes = content_bytes[i : i + bytes_per_file]

            suffix = self._generate_suffix(file_index, suffix_length, numeric_suffix)
            output_filename = f"{prefix}{suffix}"

            if verbose:
                self.console.print(f"creating file '{output_filename}'")

            # Write to remote file
            from io import BytesIO

            output_stream = BytesIO(chunk_bytes)
            client.push(output_filename, output_stream)

    def _split_by_line_bytes(
        self,
        client: ClientType,
        content: str,
        prefix: str,
        bytes_per_line: int,
        suffix_length: int,
        numeric_suffix: bool,
        verbose: bool,
    ):
        """Split content by bytes per line."""
        lines = content.splitlines(keepends=True)
        file_index = 0
        current_chunk = []
        current_size = 0

        for line in lines:
            line_size = len(line.encode("utf-8"))

            if current_size + line_size > bytes_per_line and current_chunk:
                # Save current chunk
                chunk_content = "".join(current_chunk)
                suffix = self._generate_suffix(
                    file_index, suffix_length, numeric_suffix
                )
                output_filename = f"{prefix}{suffix}"

                if verbose:
                    self.console.print(f"creating file '{output_filename}'")

                from io import StringIO

                output_stream = StringIO(chunk_content)
                client.push(output_filename, output_stream)

                file_index += 1
                current_chunk = [line]
                current_size = line_size
            else:
                current_chunk.append(line)
                current_size += line_size

        # Save final chunk if any
        if current_chunk:
            chunk_content = "".join(current_chunk)
            suffix = self._generate_suffix(file_index, suffix_length, numeric_suffix)
            output_filename = f"{prefix}{suffix}"

            if verbose:
                self.console.print(f"creating file '{output_filename}'")

            from io import StringIO

            output_stream = StringIO(chunk_content)
            client.push(output_filename, output_stream)



"""Script command for Cascade.

This module provides implementation for terminal session recording
command that works with the Pebble filesystem API.

The script command simulates the functionality of the traditional Unix script utility
while working within the Cascade shell environment.
"""

from __future__ import annotations

import json
import time
from typing import TYPE_CHECKING, Union

import ops

from ...utils.command_helpers import handle_help_flag, parse_flags, safe_read_file
from .._base import Command

if TYPE_CHECKING:
    import shimmer


# TODO: Use the prototype from Shimmer.
ClientType = Union[ops.pebble.Client, "shimmer.PebbleCliClient"]


class ScriptError(Exception):
    """Exception raised for script command errors."""


class ScriptCommand(Command):
    """Implementation of script command."""

    name = "script"
    help = "Record terminal sessions to typescript files"
    category = "Terminal"

    def show_help(self):
        """Show command help."""
        help_text = """Record terminal sessions to typescript files.

Usage: script [OPTIONS] [FILE]

Description:
    Record terminal sessions including input and output with timing
    information. Creates typescript and timing files for later replay.

Options:
    -a, --append        Append to existing typescript file
    -c COMMAND          Run specific command instead of shell
    -e, --return        Return exit code of child process
    -f, --flush         Flush output after each write
    -o, --output-limit SIZE  Maximum output per session (bytes)
    -q, --quiet         Be quiet (don't print start/stop messages)
    -t TIMINGFILE       Write timing data to file (default: timing)
    --timing-format FORMAT  Timing format: classic or advanced
    -h, --help          Show this help message

Examples:
    script                      # Record to typescript
    script session.log          # Record to specific file
    script -c "ls -la"          # Record specific command
    script -t timing.log out.log # Custom timing and output files
    script -q session.txt       # Quiet mode
        """
        self.console.print(help_text)

    def execute(self, client: ClientType, args: list[str]) -> int:
        """Execute the script command."""
        if handle_help_flag(self, args):
            return 0

        # Parse flags
        parse_result = parse_flags(
            args,
            {
                "a": bool,  # append
                "append": bool,  # append
                "c": str,  # command
                "e": bool,  # return exit code
                "return": bool,  # return exit code
                "f": bool,  # flush
                "flush": bool,  # flush
                "o": int,  # output limit
                "output-limit": int,  # output limit
                "q": bool,  # quiet
                "quiet": bool,  # quiet
                "t": str,  # timing file
                "timing-format": str,  # timing format
            },
            self.shell,
        )
        if parse_result is None:
            return 1
        flags, positional_args = parse_result

        # Determine output file
        output_file = positional_args[0] if positional_args else "typescript"

        append_mode = flags.get("a", False) or flags.get("append", False)
        command = flags.get("c")
        return_exit = flags.get("e", False) or flags.get("return", False)
        flush_output = flags.get("f", False) or flags.get("flush", False)
        output_limit = flags.get("o") or flags.get("output-limit")
        quiet = flags.get("q", False) or flags.get("quiet", False)
        timing_file = flags.get("t") or "timing"
        timing_format = flags.get("timing-format", "classic")

        try:
            return self._start_recording(
                client,
                output_file,
                timing_file,
                append_mode,
                command,
                return_exit,
                flush_output,
                output_limit,
                quiet,
                timing_format,
            )

        except Exception as e:
            self.console.print(f"[red]script: {e}[/red]")
            return 1

    def _start_recording(
        self,
        client: ClientType,
        output_file: str,
        timing_file: str,
        append_mode: bool,
        command: str | None,
        return_exit: bool,
        flush_output: bool,
        output_limit: int | None,
        quiet: bool,
        timing_format: str,
    ) -> int:
        """Start recording a terminal session."""
        if not quiet:
            if command:
                self.console.print(f"Script started, command: {command}")
            else:
                self.console.print("Script started, recording to terminal")
            self.console.print(f"Output file is {output_file}")
            if timing_file:
                self.console.print(f"Timing file is {timing_file}")

        # Check if files exist for append mode
        if append_mode:
            existing_output = safe_read_file(client, output_file)
            existing_timing = (
                safe_read_file(client, timing_file) if timing_file else None
            )

            if existing_output is not None and not quiet:
                self.console.print(f"Appending to existing file {output_file}")
            if existing_timing is not None and timing_file and not quiet:
                self.console.print(f"Appending to existing timing file {timing_file}")

        # Simulate recording session
        start_time = time.time()
        session_data = []
        timing_data = []

        if command:
            # Record specific command execution
            exit_code = self._record_command(
                client,
                command,
                session_data,
                timing_data,
                start_time,
                output_limit,
                flush_output,
            )
        else:
            # Record interactive session simulation
            exit_code = self._record_interactive_session(
                client,
                session_data,
                timing_data,
                start_time,
                output_limit,
                flush_output,
                quiet,
            )

        # Generate output content
        session_output = self._format_session_output(session_data, start_time)

        # Generate timing content
        timing_output = ""
        if timing_file:
            timing_output = self._format_timing_output(
                timing_data, timing_format, start_time
            )

        # Show what would be written (since we're in read-only mode)
        mode = "append to" if append_mode else "write to"

        self.console.print(f"[yellow]script: would {mode} {output_file}:[/yellow]")
        if len(session_output) > 1000:
            self.console.print(f"[dim]{session_output[:500]}...[/dim]")
            self.console.print(
                f"[dim]...(truncated, total {len(session_output)} bytes)[/dim]"
            )
        else:
            self.console.print(f"[dim]{session_output}[/dim]")

        if timing_file:
            self.console.print(f"[yellow]script: would {mode} {timing_file}:[/yellow]")
            if len(timing_output) > 500:
                self.console.print(f"[dim]{timing_output[:200]}...[/dim]")
                self.console.print(
                    f"[dim]...(truncated, total {len(timing_output.splitlines())} lines)[/dim]"
                )
            else:
                self.console.print(f"[dim]{timing_output}[/dim]")

        if not quiet:
            self.console.print("Script done")
            if timing_file:
                self.console.print(f"Timing data saved to {timing_file}")

        return exit_code if return_exit else 0

    def _record_command(
        self,
        client: ClientType,
        command: str,
        session_data: list,
        timing_data: list,
        start_time: float,
        output_limit: int | None,
        flush_output: bool,
    ) -> int:
        """Record execution of a specific command."""
        current_time = time.time()

        # Record command input
        session_data.append(
            {"timestamp": current_time, "type": "input", "data": f"{command}\r\n"}
        )
        timing_data.append(
            {
                "timestamp": current_time - start_time,
                "type": "input",
                "size": len(command) + 2,
            }
        )

        # TODO: Finish the actual command execution.
        try:
            output = f"[Simulated output of command: {command}]\n"
            output += "This would be the actual command output in a real session.\n"

            current_time = time.time()
            session_data.append(
                {"timestamp": current_time, "type": "output", "data": output}
            )
            timing_data.append(
                {
                    "timestamp": current_time - start_time,
                    "type": "output",
                    "size": len(output),
                }
            )

            # Check output limit
            if output_limit and len(output) > output_limit:
                self.console.print(
                    f"[yellow]script: output limit of {output_limit} bytes exceeded[/yellow]"
                )
                return 1

            return 0

        except Exception:
            return 1

    def _record_interactive_session(
        self,
        client: ClientType,
        session_data: list,
        timing_data: list,
        start_time: float,
        output_limit: int | None,
        flush_output: bool,
        quiet: bool,
    ) -> int:
        """Record an interactive session simulation."""
        if not quiet:
            self.console.print(
                "[yellow]script: simulating interactive session[/yellow]"
            )
            self.console.print(
                "[dim]In a real implementation, this would start an interactive shell[/dim]"
            )
        # TODO: Finish this implementation.
        return 0

    def _format_session_output(self, session_data: list, start_time: float) -> str:
        """Format session data into typescript format."""
        output_lines = []
        output_lines.append(f"Script started on {time.ctime(start_time)}")

        output_lines.extend(entry["data"] for entry in session_data)

        output_lines.append(f"Script done on {time.ctime()}")
        return "".join(output_lines)

    def _format_timing_output(
        self, timing_data: list, timing_format: str, start_time: float
    ) -> str:
        """Format timing data."""
        if timing_format == "advanced":
            # Advanced format with JSON
            return json.dumps(
                {"version": 2, "start_time": start_time, "events": timing_data},
                indent=2,
            )
        else:
            # Classic format: timestamp size
            lines = []
            for entry in timing_data:
                timestamp = entry["timestamp"]
                size = entry["size"]
                lines.append(f"{timestamp:.6f} {size}")
            return "\n".join(lines) + "\n"



"""Implementation of HostnameCommand."""

from __future__ import annotations

from typing import TYPE_CHECKING, Union

import ops

from ...utils.command_helpers import handle_help_flag, parse_flags, safe_read_file
from ...utils.proc_reader import read_proc_file
from ...utils.theme import get_theme
from .._base import Command
from .exceptions import SystemInfoError

if TYPE_CHECKING:
    import shimmer


# TODO: Use the prototype from Shimmer.
ClientType = Union[ops.pebble.Client, "shimmer.PebbleCliClient"]


class HostnameCommand(Command):
    """Implementation of hostname command."""

    name = "hostname"
    help = "Display the system hostname"
    category = "System Information"

    def show_help(self):
        """Show command help."""
        help_text = """Display the system hostname.

Usage: hostname [OPTIONS]

Options:
    -h, --help      Show this help message
    -f, --fqdn      Display the fully qualified domain name
    -i, --ip        Display the IP address
    -I, --all-ip    Display all IP addresses
    -A, --all-fqdn  Display all FQDNs
    -s, --short     Display the short hostname
    -d, --domain    Display the domain name
        """
        self.console.print(help_text)

    def execute(self, client: ClientType, args: list[str]) -> int:
        """Execute the hostname command."""
        self.client = client
        if handle_help_flag(self, args):
            return 0

        # Parse flags
        parse_result = parse_flags(
            args,
            {
                "f": bool,  # fully qualified domain name
                "s": bool,  # short hostname
                "d": bool,  # domain name
                "i": bool,  # IP address
                "I": bool,  # all IP addresses
                "A": bool,  # all FQDNs
            },
            self.shell,
        )
        if parse_result is None:
            return 1
        flags, positional_args = parse_result

        fqdn = flags.get("f", False)
        short = flags.get("s", False)
        domain = flags.get("d", False)
        ip_address = flags.get("i", False)
        all_ips = flags.get("I", False)
        all_fqdns = flags.get("A", False)

        try:
            if ip_address or all_ips:
                return self._show_ip_addresses(all_ips)
            elif domain:
                return self._show_domain()
            elif all_fqdns:
                return self._show_all_fqdns()
            elif fqdn:
                return self._show_fqdn()
            elif short:
                return self._show_short_hostname()
            else:
                # Default behavior - show hostname
                return self._show_hostname()

        except SystemInfoError as e:
            self.console.print(get_theme().error_text(f"hostname: {e}"))
            return 1

    def _get_help_text(self) -> str:
        """Get help text for the hostname command."""
        return """
hostname - display or set system hostname

USAGE:
    hostname [OPTIONS]

OPTIONS:
    -f, --fqdn      Display fully qualified domain name
    -s, --short     Display short hostname (default)
    -d, --domain    Display domain name only
    -i, --ip-address Display IP address
    -I, --all-ip-addresses Display all IP addresses
    -A, --all-fqdns Display all FQDNs
    -h, --help      Show this help message

EXAMPLES:
    hostname        # Show hostname
    hostname -f     # Show FQDN
    hostname -i     # Show IP address
"""

    def _show_hostname(self) -> int:
        """Show the system hostname."""
        try:
            # Try to read hostname from /etc/hostname
            try:
                hostname_content = safe_read_file(self.client, "/etc/hostname")
                hostname = hostname_content.strip() if hostname_content else ""
                if hostname:
                    self.console.print(hostname)
                    return 0
            except ops.pebble.PathError:
                pass

            # Fallback: try to get hostname from /proc/sys/kernel/hostname
            try:
                with self.client.pull(
                    "/proc/sys/kernel/hostname", encoding="utf-8"
                ) as f:
                    hostname = f.read().strip()
                if hostname:
                    self.console.print(hostname)
                    return 0
            except ops.pebble.PathError:
                pass

            # Last resort: check uname for nodename
            hostname = self._get_uname_field("nodename")
            if hostname:
                self.console.print(hostname)
                return 0

            raise SystemInfoError("Cannot determine hostname")
        except Exception as e:
            raise SystemInfoError(f"Failed to get hostname: {e}") from e

    def _show_short_hostname(self) -> int:
        """Show short hostname."""
        try:
            hostname = self._get_hostname()
            short_name = hostname.split(".")[0]
            self.console.print(short_name)
            return 0
        except Exception as e:
            raise SystemInfoError(f"Failed to get short hostname: {e}") from e

    def _show_fqdn(self) -> int:
        """Show fully qualified domain name."""
        try:
            # Try to resolve FQDN using hostname and /etc/hosts
            hostname = self._get_hostname()

            # If hostname already contains dots, it might be FQDN
            if "." in hostname:
                self.console.print(hostname)
                return 0

            # Try to find FQDN in /etc/hosts
            try:
                hosts_content = safe_read_file(self.client, "/etc/hosts") or ""

                for line in hosts_content.splitlines():
                    line = line.strip()
                    if line and not line.startswith("#"):
                        parts = line.split()
                        if len(parts) >= 2:
                            for name in parts[1:]:
                                if name.startswith(hostname + "."):
                                    self.console.print(name)
                                    return 0
            except ops.pebble.PathError:
                pass

            # Fallback: just return hostname
            self.console.print(hostname)
            return 0
        except Exception as e:
            raise SystemInfoError(f"Failed to get FQDN: {e}") from e

    def _show_domain(self) -> int:
        """Show domain name only."""
        try:
            fqdn = self._get_hostname()
            parts = fqdn.split(".", 1)
            if len(parts) > 1:
                self.console.print(parts[1])
            else:
                self.console.print("")
            return 0
        except Exception as e:
            raise SystemInfoError(f"Failed to get domain: {e}") from e

    def _show_ip_addresses(self, show_all: bool) -> int:
        """Show IP address(es)."""
        try:
            # Read network interfaces from /proc/net/fib_trie or /proc/net/route
            # This is complex, so for now we'll use a simpler approach

            # Try to get IP from hostname resolution in /etc/hosts
            hostname = self._get_hostname()

            try:
                hosts_content = safe_read_file(self.client, "/etc/hosts") or ""

                ips = set()
                for line in hosts_content.splitlines():
                    line = line.strip()
                    if line and not line.startswith("#"):
                        parts = line.split()
                        if len(parts) >= 2 and hostname in parts[1:]:
                            ip = parts[0]
                            # Simple IP validation
                            if self._is_valid_ip(ip):
                                ips.add(ip)

                if ips:
                    if show_all:
                        for ip in sorted(ips):
                            self.console.print(ip)
                    else:
                        # Show first non-localhost IP
                        for ip in sorted(ips):
                            if ip != "127.0.0.1" and ip != "::1":
                                self.console.print(ip)
                                return 0
                        # Fallback to localhost
                        self.console.print(next(iter(ips)))
                    return 0
            except ops.pebble.PathError:
                pass

            # Fallback: show localhost
            self.console.print("127.0.0.1")
            return 0
        except Exception as e:
            raise SystemInfoError(f"Failed to get IP address: {e}") from e

    def _show_all_fqdns(self) -> int:
        """Show all FQDNs."""
        try:
            hostname = self._get_hostname()
            fqdns = set([hostname])

            # Look in /etc/hosts for additional names
            try:
                hosts_content = safe_read_file(self.client, "/etc/hosts") or ""

                for line in hosts_content.splitlines():
                    line = line.strip()
                    if line and not line.startswith("#"):
                        parts = line.split()
                        if len(parts) >= 2:
                            for name in parts[1:]:
                                if "." in name:  # FQDN
                                    fqdns.add(name)
            except ops.pebble.PathError:
                pass

            for name in sorted(fqdns):
                self.console.print(name)
            return 0
        except Exception as e:
            raise SystemInfoError(f"Failed to get FQDNs: {e}") from e

    def _get_hostname(self) -> str:
        """Get hostname from remote system."""
        # Try /etc/hostname first
        try:
            hostname_content = safe_read_file(self.client, "/etc/hostname")
            hostname = hostname_content.strip() if hostname_content else ""
            if hostname:
                return hostname
        except ops.pebble.PathError:
            pass

        # Try /proc/sys/kernel/hostname
        try:
            hostname = read_proc_file(self.client, "/proc/sys/kernel/hostname").strip()
            if hostname:
                return hostname
        except ops.pebble.PathError:
            pass

        # Fallback to uname nodename
        nodename = self._get_uname_field("nodename")
        if nodename:
            return nodename

        return "localhost"

    def _is_valid_ip(self, ip: str) -> bool:
        """Simple IP address validation."""
        parts = ip.split(".")
        if len(parts) != 4:
            return False
        try:
            for part in parts:
                num = int(part)
                if num < 0 or num > 255:
                    return False
            return True
        except ValueError:
            return False

    def _get_uname_field(self, field: str) -> str:
        """Get a field from uname via /proc/version or similar."""
        try:
            # Try to read from /proc/sys/kernel/ files
            if field == "nodename":
                try:
                    with self.client.pull(
                        "/proc/sys/kernel/hostname", encoding="utf-8"
                    ) as f:
                        return f.read().strip()
                except ops.pebble.PathError:
                    pass
            elif field == "release":
                try:
                    with self.client.pull(
                        "/proc/sys/kernel/osrelease", encoding="utf-8"
                    ) as f:
                        return f.read().strip()
                except ops.pebble.PathError:
                    pass
            elif field == "version":
                try:
                    with self.client.pull("/proc/version", encoding="utf-8") as f:
                        version_line = f.read().strip()
                        # Extract version info from the version string
                        return version_line
                except ops.pebble.PathError:
                    pass
        except Exception:  # noqa: S110
            # Broad exception handling needed for system info gathering
            pass

        return ""



"""Integration tests for the YqCommand."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

import pebble_shell.commands
import pebble_shell.shell

if TYPE_CHECKING:
    import ops


@pytest.fixture
def command(shell: pebble_shell.shell.PebbleShell):
    """Fixture to create a YqCommand instance."""
    yield pebble_shell.commands.YqCommand(shell=shell)


def test_name(command: pebble_shell.commands.YqCommand):
    assert command.name == "yq <file> [keypath]"


def test_category(command: pebble_shell.commands.YqCommand):
    assert command.category == "Built-in Commands"


def test_help(command: pebble_shell.commands.YqCommand):
    with command.shell.console.capture() as capture:
        command.show_help()
    assert "yq" in capture.get()


@pytest.mark.parametrize("args", [["-h"], ["--help"]])
def test_execute_help(
    client: ops.pebble.Client,
    command: pebble_shell.commands.YqCommand,
    args: list[str],
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=args)
    assert result == 0
    assert "yq" in capture.get()


def test_execute_no_args(
    client: ops.pebble.Client,
    command: pebble_shell.commands.YqCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])
    # Should fail with usage message
    assert result == 1
    output = capture.get()
    assert "No file specified" in output
    assert "Usage: yq <file> [.foo.bar]" in output


def test_execute_nonexistent_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.YqCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/nonexistent/file.yaml"])
    # Should fail with file error
    assert result == 1
    output = capture.get()
    assert "Error reading file" in output


def test_execute_invalid_yaml_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.YqCommand,
):
    # Try to parse a non-YAML file like /etc/passwd
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    # /etc/passwd might be valid YAML (plain text is often valid YAML)
    # but we expect it to succeed since YAML is more permissive than JSON
    if result == 1:
        output = capture.get()
        assert "Error parsing YAML" in output
    else:
        # If it succeeds, should show YAML output
        assert result == 0


def test_execute_with_valid_yaml_like_file(
    client: ops.pebble.Client,
    command: pebble_shell.commands.YqCommand,
):
    # /etc/passwd might be parsed as YAML (text files can be valid YAML)
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    # Should either succeed (if treated as valid YAML) or fail with parsing error
    if result == 0:
        # Should show YAML output formatting
        output = capture.get()
        assert "yq Output" in output
    else:
        assert result == 1
        output = capture.get()
        assert "Error parsing YAML" in output


def test_execute_with_keypath_invalid_format(
    client: ops.pebble.Client,
    command: pebble_shell.commands.YqCommand,
):
    # Test invalid keypath format (not starting with dot)
    with command.shell.console.capture() as _:
        result = command.execute(client=client, args=["/etc/passwd", "foo.bar"])

    # Should fail with keypath format error (if YAML parsing succeeds)
    # or fail with YAML parsing error first
    assert result == 1


def test_execute_with_valid_keypath_format(
    client: ops.pebble.Client,
    command: pebble_shell.commands.YqCommand,
):
    # Test valid keypath format (starting with dot)
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd", ".foo.bar"])

    # Should fail with key path error (assuming /etc/passwd doesn't have 'foo' key)
    # or succeed if the file is parsed as YAML and has the key
    if result == 1:
        output = capture.get()
        # Could fail on YAML parsing or key path navigation
        assert any(msg in output for msg in ["Error parsing YAML", "Key path error"])


def test_execute_keypath_navigation_dict_access(
    client: ops.pebble.Client,
    command: pebble_shell.commands.YqCommand,
):
    # Test dictionary key access in keypath
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd", ".users.0.name"])

    # Should fail with key path error (assuming structure doesn't exist)
    if result == 1:
        output = capture.get()
        assert any(msg in output for msg in ["Error parsing YAML", "Key path error"])


def test_execute_keypath_navigation_array_access(
    client: ops.pebble.Client,
    command: pebble_shell.commands.YqCommand,
):
    # Test array index access in keypath
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd", ".items.0"])

    # Should fail with key path error (assuming structure doesn't exist)
    if result == 1:
        output = capture.get()
        assert any(msg in output for msg in ["Error parsing YAML", "Key path error"])


def test_execute_empty_keypath(
    client: ops.pebble.Client,
    command: pebble_shell.commands.YqCommand,
):
    # Test empty keypath (just dot)
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd", "."])

    # Should succeed if YAML parsing succeeds (empty keypath returns whole document)
    if result == 0:
        output = capture.get()
        assert "yq Output" in output
    else:
        assert result == 1


def test_execute_invalid_keypath_missing_dot(
    client: ops.pebble.Client,
    command: pebble_shell.commands.YqCommand,
):
    # Test keypath that doesn't start with dot
    with command.shell.console.capture() as _:
        result = command.execute(client=client, args=["/etc/passwd", "foo"])

    # Should fail with invalid keypath error
    assert result == 1
    # Note: May fail on YAML parsing first, but if YAML is valid:
    # Should show keypath validation error


def test_execute_multiple_args_extra_ignored(
    client: ops.pebble.Client,
    command: pebble_shell.commands.YqCommand,
):
    # Test with more than 2 arguments (extras should be ignored)
    with command.shell.console.capture() as capture:
        result = command.execute(
            client=client, args=["/etc/passwd", ".foo", "extra", "args"]
        )

    # Should use first file and first keypath, ignore extras
    if result == 1:
        output = capture.get()
        assert any(msg in output for msg in ["Error parsing YAML", "Key path error"])


def test_execute_path_resolution_relative(
    client: ops.pebble.Client,
    command: pebble_shell.commands.YqCommand,
):
    # Test relative path resolution
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["../etc/passwd"])

    # Should resolve relative path and then parse as YAML
    if result == 0:
        output = capture.get()
        assert "yq Output" in output
    else:
        assert result == 1


def test_execute_path_resolution_absolute(
    client: ops.pebble.Client,
    command: pebble_shell.commands.YqCommand,
):
    # Test absolute path
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    # Should handle absolute path and parse as YAML
    if result == 0:
        output = capture.get()
        assert "yq Output" in output
    else:
        assert result == 1


def test_execute_yaml_parsing_with_plain_text(
    client: ops.pebble.Client,
    command: pebble_shell.commands.YqCommand,
):
    # YAML is more permissive than JSON - plain text can be valid YAML
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    # May succeed since YAML can parse plain text
    if result == 0:
        output = capture.get()
        assert "yq Output" in output
    else:
        assert result == 1
        output = capture.get()
        assert "Error parsing YAML" in output


def test_execute_output_formatting_with_valid_yaml(
    client: ops.pebble.Client,
    command: pebble_shell.commands.YqCommand,
):
    # Test the output formatting logic
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    if result == 0:
        # Should show formatted YAML output with syntax highlighting
        output = capture.get()
        assert "yq Output" in output
    else:
        # Should show error in formatted panel
        assert result == 1


def test_execute_syntax_highlighting_preparation(
    client: ops.pebble.Client,
    command: pebble_shell.commands.YqCommand,
):
    # Test that the command prepares for syntax highlighting
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    if result == 0:
        output = capture.get()
        # Should have YAML syntax highlighting
        assert "yq Output" in output
    else:
        # Should show proper error formatting
        output = capture.get()
        assert "yq Error" in output


def test_execute_file_read_error_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.YqCommand,
):
    # Test file reading error handling
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/nonexistent/path/file.yaml"])

    # Should fail with file reading error
    assert result == 1
    output = capture.get()
    assert "Error reading file" in output


def test_execute_yaml_parsing_error_details(
    client: ops.pebble.Client,
    command: pebble_shell.commands.YqCommand,
):
    # Test that YAML parsing errors provide details
    # This is harder to trigger since YAML is very permissive
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    # Should either succeed or show detailed error
    if result == 1:
        output = capture.get()
        assert "yq Error" in output


def test_execute_keypath_error_handling(
    client: ops.pebble.Client,
    command: pebble_shell.commands.YqCommand,
):
    # Test keypath validation error handling
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd", "invalid_keypath"])

    # Should fail with keypath validation error (if YAML parsing succeeds)
    if result == 1:
        output = capture.get()
        # Could be YAML parsing error or keypath validation error
        assert "yq Error" in output


def test_execute_yaml_safe_dump_formatting(
    client: ops.pebble.Client,
    command: pebble_shell.commands.YqCommand,
):
    # Test YAML output formatting with safe_dump
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["/etc/passwd"])

    if result == 0:
        # Should use yaml.safe_dump for output formatting
        output = capture.get()
        assert "yq Output" in output
    else:
        assert result == 1



"""Integration tests for the SortCommand."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

import pebble_shell.commands
import pebble_shell.shell

if TYPE_CHECKING:
    import ops


@pytest.fixture
def command(shell: pebble_shell.shell.PebbleShell):
    """Fixture to create a SortCommand instance."""
    yield pebble_shell.commands.SortCommand(shell=shell)


def test_name(command: pebble_shell.commands.SortCommand):
    assert command.name == "sort [-r] [-n] [-k field] <file> [file2...]"


def test_category(command: pebble_shell.commands.SortCommand):
    assert command.category == "Filesystem Commands"


def test_help(command: pebble_shell.commands.SortCommand):
    with command.shell.console.capture() as capture:
        command.show_help()
    assert "Sort lines in files" in capture.get()


@pytest.mark.parametrize("args", [["-h"], ["--help"]])
def test_execute_help(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SortCommand,
    args: list[str],
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=args)
    assert result == 0
    assert "Sort lines in files" in capture.get()


def test_execute_no_args(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SortCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=[])
    assert result == 1
    assert "Usage: sort [-r] [-n] [-k field] <file>" in capture.get()


def test_execute_invalid_field_number(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SortCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-k", "invalid", "file"])
    assert result == 1
    assert "Error: Flag -k requires an integer argument" in capture.get()


def test_execute_missing_k_argument(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SortCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-k"])
    assert result == 1
    assert "Error: Flag -k requires an argument" in capture.get()


def test_execute_invalid_option(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SortCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-x", "file"])
    assert result == 1
    assert "Error: Invalid option -x" in capture.get()


def test_execute_no_input_files(
    client: ops.pebble.Client,
    command: pebble_shell.commands.SortCommand,
):
    with command.shell.console.capture() as capture:
        result = command.execute(client=client, args=["-r"])
    assert result == 1
    assert "Usage: sort [-r] [-n] [-k field] <file>" in capture.get()
